# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter endpoint capability wrapper."""
import copy
from typing import Any, Callable, Collection, List, Mapping, Optional, Set, Type

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_endpoints_and_clusters
from gazoo_device.capabilities.interfaces import matter_endpoints_base
from gazoo_device.capabilities.matter_clusters.interfaces import cluster_base
from gazoo_device.capabilities.matter_endpoints import unsupported_endpoint
from gazoo_device.capabilities.matter_endpoints.interfaces import endpoint_base
from gazoo_device.protos import attributes_service_pb2
from gazoo_device.protos import descriptor_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport
from gazoo_device.utility import pwrpc_utils

_DESCRIPTOR_SERVICE_NAME = "Descriptor"
_DESCRIPTOR_GET_ENDPOINTS_RPC_NAME = "PartsList"
_DESCRIPTOR_DEVICE_TYPE_RPC_NAME = "DeviceTypeList"
_DESCRIPTOR_GET_CLUSTERS_RPC_NAME = "ServerList"
_ATTRIBUTE_SERVICE_NAME = "Attributes"
_ATTRIBUTE_READ_RPC_NAME = "Read"
_ATTRIBUTE_WRITE_RPC_NAME = "Write"
_ATTRIBUTE_DATA_MODULE_PATH = "gazoo_device.protos.attributes_service_pb2.AttributeData"
_ATTRIBUTE_METADATA_MODULE_PATH = "gazoo_device.protos.attributes_service_pb2.AttributeMetadata"

logger = gdm_logger.get_logger()


class _DescriptorServiceHandler:
  """Descriptor RPC service handler.

  Handler which leverages the descriptor RPC service on Matter device to get
  the list of supported endpoint IDs and their device types. It allows the
  implementation of generic Matter device controller on different platforms.
  """

  def __init__(
      self,
      device_name: str,
      switchboard_call: Callable[..., Any],
      rpc_timeout_s: int):
    """Creates a descriptor service handler instance.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The switchboard.call method.
      rpc_timeout_s: Timeout (s) for RPC calls.
    """
    self._switchboard_call = switchboard_call
    self._rpc_timeout_s = rpc_timeout_s
    self._device_name = device_name

    # The endpoint ID to endpoint class mapping.
    self._endpoint_id_to_class = {}

    # The endpoint class to endpoint ID mapping.
    self._endpoint_class_to_id = {}

    # The endpoint ID to set of cluster classes mapping.
    self._endpoint_id_to_clusters = {}

    # The endpoint ID to device type ID mapping.
    self._endpoint_id_to_device_type_id = {}

  @property
  def endpoint_id_to_class(
      self) -> Mapping[int, Optional[Type[endpoint_base.EndpointBase]]]:
    """Returns the endpoint_id_to_class mapping."""
    self._fetch_endpoints_and_clusters()
    return copy.deepcopy(self._endpoint_id_to_class)

  @property
  def endpoint_class_to_id(
      self) -> Mapping[Type[endpoint_base.EndpointBase], int]:
    """Returns the endpoint_class_to_id mapping."""
    self._fetch_endpoints_and_clusters()
    return copy.deepcopy(self._endpoint_class_to_id)

  @property
  def endpoint_id_to_clusters(
      self) -> Mapping[int, Set[Type[cluster_base.ClusterBase]]]:
    """Returns the endpoint ID to cluster classes mapping."""
    self._fetch_endpoints_and_clusters()
    return copy.deepcopy(self._endpoint_id_to_clusters)

  @property
  def endpoint_id_to_device_type_id(self) -> Mapping[int, int]:
    """Returns the endpoint ID to device type ID mapping."""
    self._fetch_endpoints_and_clusters()
    return copy.deepcopy(self._endpoint_id_to_device_type_id)

  def reset(self) -> None:
    """Resets the endpoint ID and endpoint class mapping."""
    self._endpoint_id_to_class.clear()
    self._endpoint_class_to_id.clear()
    self._endpoint_id_to_clusters.clear()
    self._endpoint_id_to_device_type_id.clear()

  def _fetch_endpoints_and_clusters(self) -> None:
    """Retrieves the supported endpoints from the descriptor RPC service.

    The set of supported clusters is also obtained by the descriptor RPC. Note
    that RPC is only triggered when endpoint_id_to_class mapping is None.
    """
    if not self._endpoint_id_to_class:
      for endpoint_id in self._get_supported_endpoint_ids():
        endpoint_cls = self._get_endpoint_class(endpoint_id)
        self._endpoint_id_to_class[endpoint_id] = endpoint_cls
        # Ensuring we store the first endpoint ID handled by this class.
        # This mapping will be used in get_endpoint_instance_by_class method
        if endpoint_cls not in self._endpoint_class_to_id:
          self._endpoint_class_to_id[endpoint_cls] = endpoint_id
        self._endpoint_id_to_clusters[endpoint_id] = (
            self._get_supported_clusters(endpoint_id))

  def _get_supported_endpoint_ids(self) -> List[int]:
    """Gets the list of supported endpoint ids on the device."""
    ack, list_of_supported_endpoints = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_DESCRIPTOR_SERVICE_NAME,
                     _DESCRIPTOR_GET_ENDPOINTS_RPC_NAME),
        method_kwargs={
            "endpoint": matter_endpoints_base.ROOT_NODE_ENDPOINT_ID,
            "pw_rpc_timeout_s": self._rpc_timeout_s})
    if not ack:
      raise errors.DeviceError(
          f"Device {self._device_name} getting {_DESCRIPTOR_SERVICE_NAME} "
          f"{_DESCRIPTOR_GET_ENDPOINTS_RPC_NAME} failed.")
    supported_endpoint_ids = []
    for endpoint_in_bytes in list_of_supported_endpoints:
      endpoint = descriptor_service_pb2.Endpoint.FromString(endpoint_in_bytes)
      supported_endpoint_ids.append(endpoint.endpoint)
    return supported_endpoint_ids

  def _get_endpoint_class(
      self, endpoint_id: int) -> Type[endpoint_base.EndpointBase]:
    """Gets the endpoint class by the given endpoint id.

    Args:
      endpoint_id: The given endpoint ID on the device.

    Returns:
      The endpoint class module. The method returns None if the given endpoint
      ID does not have device type.
    """
    ack, device_types = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_DESCRIPTOR_SERVICE_NAME,
                     _DESCRIPTOR_DEVICE_TYPE_RPC_NAME),
        method_kwargs={
            "endpoint": endpoint_id, "pw_rpc_timeout_s": self._rpc_timeout_s})
    if not ack:
      raise errors.DeviceError(
          f"Device {self._device_name} getting {_DESCRIPTOR_SERVICE_NAME} "
          f"{_DESCRIPTOR_DEVICE_TYPE_RPC_NAME} failed.")
    device_type = descriptor_service_pb2.DeviceType.FromString(device_types[0])
    device_type_id = device_type.device_type

    # Store the endpoint ID to device type ID mapping.
    self._endpoint_id_to_device_type_id[endpoint_id] = device_type_id

    return matter_endpoints_and_clusters.MATTER_DEVICE_TYPE_ID_TO_CLASS.get(
        device_type_id, unsupported_endpoint.UnsupportedEndpoint)

  def _get_supported_clusters(
      self, endpoint_id: int) -> Set[Type[cluster_base.ClusterBase]]:
    """Retrieves the supported clusters from the given endpoint ID."""
    ack, clusters = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_DESCRIPTOR_SERVICE_NAME,
                     _DESCRIPTOR_GET_CLUSTERS_RPC_NAME),
        method_kwargs={
            "endpoint": endpoint_id, "pw_rpc_timeout_s": self._rpc_timeout_s})
    if not ack:
      raise errors.DeviceError(
          f"Device {self._device_name} getting {_DESCRIPTOR_SERVICE_NAME} "
          f"{_DESCRIPTOR_GET_CLUSTERS_RPC_NAME} failed.")
    cluster_classes = set()
    for cluster_in_bytes in clusters:
      cluster = descriptor_service_pb2.Cluster.FromString(cluster_in_bytes)
      cluster_class = matter_endpoints_and_clusters.CLUSTER_ID_TO_CLASS.get(
          cluster.cluster_id)
      if cluster_class is None:
        logger.warning(
            f"Cluster class for cluster ID {hex(cluster.cluster_id)} has not "
            "been implemented yet.")
        continue
      cluster_classes.add(cluster_class)
    return cluster_classes


class MatterEndpointsAccessor(matter_endpoints_base.MatterEndpointsBase):
  """Capability wrapper for accessing the Matter endpoint instances."""

  def __init__(
      self,
      device_name: str,
      switchboard_call: Callable[..., Any],
      rpc_timeout_s: int):
    """Constructor of MatterEndpointsAccessor.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The switchboard.call method.
      rpc_timeout_s: Timeout (s) for RPC calls.
    """
    super().__init__(device_name=device_name)

    # DescriptorServiceHandler instance for accessing endpoints and clusters
    # information via Matter descriptor cluster.
    self._descriptor_service_handler = _DescriptorServiceHandler(
        device_name=device_name,
        switchboard_call=switchboard_call,
        rpc_timeout_s=rpc_timeout_s)

    # Endpoint ID to endpoint instance mapping
    self._endpoints = {}

    # Arguments for Ember APIs.
    self._switchboard_call = switchboard_call
    self._rpc_timeout_s = rpc_timeout_s

  @decorators.CapabilityLogDecorator(logger)
  def get(self, endpoint_id: int) -> endpoint_base.EndpointBase:
    """Gets the specific endpoint instance by ID.

    Args:
      endpoint_id: Endpoint ID on the device.

    Returns:
      The endpoint class for the given endpoint ID.

    Raises:
      DeviceError: The given endpoint ID does not exist on the device. Or the
      endpoint class for the given endpoint ID is not implemented yet.
    """
    if endpoint_id not in self._endpoints:

      if (endpoint_id not in
          self._descriptor_service_handler.endpoint_id_to_class):
        raise errors.DeviceError(
            f"Endpoint ID {endpoint_id} on {self._device_name} does not exist.")

      # Obtain endpoint information from the descriptor handler.
      endpoint_class = (self._descriptor_service_handler.
                        endpoint_id_to_class[endpoint_id])

      supported_clusters = (self._descriptor_service_handler.
                            endpoint_id_to_clusters[endpoint_id])

      device_type_id = (self._descriptor_service_handler.
                        endpoint_id_to_device_type_id[endpoint_id])

      # Create endpoint instance.
      self._endpoints[endpoint_id] = endpoint_class(
          device_name=self._device_name,
          identifier=endpoint_id,
          device_type_id=device_type_id,
          supported_clusters=frozenset(supported_clusters),
          read=self.read,
          write=self.write)

    return self._endpoints[endpoint_id]

  @decorators.CapabilityLogDecorator(logger)
  def list(self) -> Mapping[int, Type[endpoint_base.EndpointBase]]:
    """Returns a mapping of endpoint ID to the supported endpoint class."""
    return self._descriptor_service_handler.endpoint_id_to_class

  @decorators.CapabilityLogDecorator(logger)
  def get_endpoint_instance_by_class(
      self, endpoint_class: Type[endpoint_base.EndpointBase]
  ) -> endpoint_base.EndpointBase:
    """Gets the endpoint instance by the given endpoint class.

    Args:
      endpoint_class: The given Matter endpoint class.

    Raises:
      DeviceError: When the given endpoint class is not supported on the device.

    Returns:
      The endpoint instance.
    """
    if (endpoint_class not in
        self._descriptor_service_handler.endpoint_class_to_id):
      raise errors.DeviceError(
          f"Class {endpoint_class} is not supported on {self._device_name}.")
    endpoint_id = (
        self._descriptor_service_handler.endpoint_class_to_id[endpoint_class])
    return self.get(endpoint_id)

  @decorators.CapabilityLogDecorator(logger)
  def reset(self) -> None:
    """Resets the endpoint ID and endpoint class mapping."""
    self._descriptor_service_handler.reset()

  @decorators.CapabilityLogDecorator(logger)
  def has_endpoints(self, endpoint_names: Collection[str]) -> bool:
    """Checks whether the device supports all the given endpoint names.

    Args:
      endpoint_names: The collection of endpoint names. The names are
        case-insensitive. Some valid examples are: "on_off_light",
        "On_Off_Light".

    Raises:
      ValueError when the given endpoint name is invalid or not supported in
      GDM.

    Returns:
      True if the device supports all the endpoints, false otherwise.
    """
    valid_endpoint_name_to_class = {
        endpoint_class.get_capability_name(): endpoint_class
        for endpoint_class in matter_endpoints_and_clusters.SUPPORTED_ENDPOINTS}
    supported_endpoints = set(self.list().values())
    for endpoint_name in endpoint_names:
      endpoint = valid_endpoint_name_to_class.get(endpoint_name.lower())
      if endpoint is None:
        raise ValueError(f"Endpoint {endpoint_name} is not recognized. "
                         "Valid endpoints are: "
                         f"{list(valid_endpoint_name_to_class.keys())}")
      if endpoint not in supported_endpoints:
        return False
    return True

  @decorators.CapabilityLogDecorator(logger)
  def get_supported_endpoints(self) -> List[str]:
    """Returns names of endpoints supported by the device."""
    return sorted(endpoint.get_capability_name() for endpoint in
                  self.get_supported_endpoint_flavors())

  @decorators.CapabilityLogDecorator(logger)
  def get_supported_endpoint_flavors(
      self) -> List[Type[endpoint_base.EndpointBase]]:
    """Returns flavors of endpoints supported by the device."""
    return [
        endpoint for endpoint in self.list().values() if endpoint is not None]

  @decorators.CapabilityLogDecorator(logger)
  def get_supported_endpoints_and_clusters(self) -> Mapping[int, Set[str]]:
    """Returns the supported endpoint IDs and set of cluster names mapping."""
    return {
        endpoint_id: self.get(endpoint_id).get_supported_clusters()
        for endpoint_id in self.list()}

  @decorators.CapabilityLogDecorator(logger)
  def get_supported_endpoint_instances_and_cluster_flavors(self) -> Mapping[
      Type[endpoint_base.EndpointBase], Set[Type[cluster_base.ClusterBase]]]:
    """Returns the supported endpoint instance and cluster flavors mapping."""
    mapping = {}
    for endpoint_id in self.list():
      endpoint = self.get(endpoint_id)
      mapping[endpoint] = endpoint.get_supported_cluster_flavors()
    return mapping

  @decorators.CapabilityLogDecorator(logger)
  def read(
      self,
      endpoint_id: int,
      cluster_id: attributes_service_pb2.ClusterType,
      attribute_id: int,
      attribute_type: attributes_service_pb2.AttributeType
  ) -> attributes_service_pb2.AttributeData:
    """Ember API read method.

    Reads attribute data from the given endpoint ID, cluster ID and
    attribute ID with the given attribute type. The endpoint ID is retrieved
    via descriptor cluster. The attribute type is defined in the
    attributes_service.proto, while cluster ID and attribute ID are defined in
    the Matter spec.

    Args:
      endpoint_id: Endpoint ID to read from.
      cluster_id: Cluster ID to read from.
      attribute_id: Attribute ID to read from.
      attribute_type: Attribute data type to read.

    Returns:
      Attribute data.

    Raises:
      Device error when ack value is false.
    """
    read_kwargs = {
        "endpoint": endpoint_id, "cluster": cluster_id,
        "attribute_id": attribute_id, "type": attribute_type,
        "pw_rpc_timeout_s": self._rpc_timeout_s}

    ack, data_in_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_ATTRIBUTE_SERVICE_NAME, _ATTRIBUTE_READ_RPC_NAME),
        method_kwargs=read_kwargs)
    if not ack:
      error_message = (
          f"Device {self._device_name} reading attribute (endpoint ID = "
          f"{endpoint_id}, cluster ID = {cluster_id}, attribute ID = "
          f"{attribute_id}) with attribute type {attribute_type} failed.")
      raise errors.DeviceError(error_message)

    return attributes_service_pb2.AttributeData.FromString(data_in_bytes)

  @decorators.CapabilityLogDecorator(logger)
  def write(
      self,
      endpoint_id: int,
      cluster_id: attributes_service_pb2.ClusterType,
      attribute_id: int,
      attribute_type: attributes_service_pb2.AttributeType,
      **data_kwargs: Any) -> None:
    """Ember API write method.

    Write attribute data to the given endpoint ID, cluster ID and
    attribute ID with the given attribute type. The endpoint ID is retrieved
    via descriptor cluster. The attribute type is defined in the
    attributes_service.proto, while cluster ID and attribute ID are defined in
    the Matter spec. data_kwargs is the data we want to write to the device, the
    supported data types are defined in AttributeData enum of
    attributes_service.proto.

    Args:
      endpoint_id: Endpoint ID to write to.
      cluster_id: Cluster ID to write to.
      attribute_id: Attribute ID to write to.
      attribute_type: Attribute data type to write.
      **data_kwargs: Attribute data to write.

    Raises:
      Device error when ack value is false.
    """
    data = attributes_service_pb2.AttributeData(**data_kwargs)
    metadata = attributes_service_pb2.AttributeMetadata(
        endpoint=endpoint_id,
        cluster=cluster_id,
        attribute_id=attribute_id,
        type=attribute_type)

    serialized_data = pwrpc_utils.PigweedProtoState(
        data, _ATTRIBUTE_DATA_MODULE_PATH)
    serialized_metadata = pwrpc_utils.PigweedProtoState(
        metadata, _ATTRIBUTE_METADATA_MODULE_PATH)
    write_kwargs = {"data": serialized_data, "metadata": serialized_metadata}

    ack, _ = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_ATTRIBUTE_SERVICE_NAME, _ATTRIBUTE_WRITE_RPC_NAME),
        method_kwargs=write_kwargs)
    if not ack:
      error_message = (
          f"Device {self._device_name} writing data: {data} to attribute ("
          f"endpoint ID = {endpoint_id}, cluster ID = {cluster_id}, attribute "
          f"ID = {attribute_id}) with attribute type {attribute_type} failed.")
      raise errors.DeviceError(error_message)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CONFIG FILE."""
import os.path
from gazoo_device import data_types
import immutabledict

PACKAGE_PATH = os.path.dirname(os.path.abspath(__file__))

INSTALL_DIRECTORY = os.path.join(os.path.expanduser("~"), "gazoo", "gdm")
LAUNCHER_PATH = os.path.join(os.path.expanduser("~"), "gazoo", "bin", "gdm")

ADB_BIN_PATH_CONFIG = "adb_path"
SEARCHWINDOWSIZE = 2000  # Default size of search window for switchboard methods

DEFAULT_LOG_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "log")
CONFIG_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "conf")
DATA_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "data")
BACKUP_PARENT_DIRECTORY = os.path.join(CONFIG_DIRECTORY, "backup")
BOTO_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "botos")
BIN_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "bin")
KEYS_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "keys")
PTY_PROCESS_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "pty_proc")
DETOK_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "detok")
VIRTUAL_ENV_DIRECTORY = os.path.join(INSTALL_DIRECTORY, "virtual_env")

REQUIRED_FOLDERS = [
    INSTALL_DIRECTORY, CONFIG_DIRECTORY, DATA_DIRECTORY, DEFAULT_LOG_DIRECTORY,
    BACKUP_PARENT_DIRECTORY, BOTO_DIRECTORY, KEYS_DIRECTORY, DETOK_DIRECTORY
]

DEFAULT_DEVICE_FILE = os.path.join(CONFIG_DIRECTORY, "devices.json")
DEFAULT_OPTIONS_FILE = os.path.join(CONFIG_DIRECTORY, "device_options.json")
DEFAULT_TESTBEDS_FILE = os.path.join(CONFIG_DIRECTORY, "testbeds.json")
DEFAULT_GDM_CONFIG_FILE = os.path.join(CONFIG_DIRECTORY, "gdm.json")
DEFAULT_LOG_FILE = os.path.join(DEFAULT_LOG_DIRECTORY, "gdm.txt")

DEVICES_KEYS = ["devices", "other_devices"]
OPTIONS_KEYS = ["device_options", "other_device_options"]
TESTBED_KEYS = ["testbeds"]

HUB_OPTION_ATTRIBUTES = ["alias"]
DEVICE_OPTION_ATTRIBUTES = [
    "alias", "power_switch", "power_port", "usb_hub", "location", "usb_port"
]

CLASS_PROPERTY_TYPES = (str, int, dict, type(None))

_BUILT_IN_EXTENSION_PACKAGE_NAME = "gazoo_device_controllers"
KEYS = immutabledict.immutabledict({
    "raspberrypi3_ssh_key": data_types.KeyInfo(
        file_name="raspberrypi3_ssh_key",
        type=data_types.KeyType.SSH,
        package=_BUILT_IN_EXTENSION_PACKAGE_NAME),
    "raspberrypi3_ssh_key_public": data_types.KeyInfo(
        file_name="raspberrypi3_ssh_key.pub",
        type=data_types.KeyType.SSH,
        package=_BUILT_IN_EXTENSION_PACKAGE_NAME),
    "unifi_switch_ssh_key": data_types.KeyInfo(
        file_name="unifi_switch_ssh_key",
        type=data_types.KeyType.SSH,
        package=_BUILT_IN_EXTENSION_PACKAGE_NAME),
    "unifi_switch_ssh_key_public": data_types.KeyInfo(
        file_name="unifi_switch_ssh_key.pub",
        type=data_types.KeyType.SSH,
        package=_BUILT_IN_EXTENSION_PACKAGE_NAME),
})



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for Matter endpoint capability wrapper."""
import abc
from typing import Mapping, Type
from gazoo_device.capabilities.interfaces import capability_base
from gazoo_device.capabilities.matter_endpoints.interfaces import endpoint_base

ROOT_NODE_ENDPOINT_ID = 0


class MatterEndpointsBase(capability_base.CapabilityBase):
  """Capability wrapper for accessing the Matter endpoint instances."""

  @abc.abstractmethod
  def get(self, endpoint_id: int) -> endpoint_base.EndpointBase:
    """Gets the specific endpoint instance by endpoint ID."""

  @abc.abstractmethod
  def list(self) -> Mapping[int, Type[endpoint_base.EndpointBase]]:
    """Lists all supported endpoints."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device controller for ESP32 Matter device.

This device controller populates the supported Matter endpoints on the ESP32
platform by using the descriptor RPC service.
"""
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.base_classes import matter_device_base
from gazoo_device.capabilities import flash_build_esptool
from gazoo_device.capabilities import pwrpc_wifi_default

logger = gdm_logger.get_logger()

_BAUDRATE = 115200
_CHIP_TYPE = "esp32"
_DEFAULT_BOOTUP_TIMEOUT_SECONDS = 15
_FLASH_MODE = "dio"
_FLASH_FREQ = "40m"


class Esp32Matter(matter_device_base.MatterDeviceBase):
  """ESP32 Matter device controller."""
  DETECT_MATCH_CRITERIA = {
      detect_criteria.PigweedQuery.IS_MATTER: True,
      detect_criteria.PigweedQuery.MANUFACTURER_NAME:
          r"(silicon(_| )labs|1a86)",
      detect_criteria.PigweedQuery.PRODUCT_NAME:
          r"(cp2104 usb to uart bridge controller|usb single serial)",
  }
  # Button definition on ESP32 M5Stack dev board:
  # 0 --> Left button on M5Stack, Button 1
  # 1 --> Middle button on M5Stack, Button 2
  # 2 --> Right button on M5Stack, Button 3
  VALID_BUTTON_IDS = (0, 1, 2)

  DEVICE_TYPE = "esp32matter"

  @decorators.PersistentProperty
  def os(self) -> str:
    """OS of ESP32 platform."""
    return "FreeRTOS"

  @decorators.PersistentProperty
  def platform(self) -> str:
    """ESP32 platform."""
    return "ESP32"

  @decorators.CapabilityDecorator(pwrpc_wifi_default.PwRPCWifiDefault)
  def pw_rpc_wifi(self):
    """PwRPCWifi capability to send RPC commands."""
    return self.lazy_init(
        pwrpc_wifi_default.PwRPCWifiDefault,
        device_name=self.name,
        switchboard_call=self.switchboard.call)

  @decorators.CapabilityDecorator(flash_build_esptool.FlashBuildEsptool)
  def flash_build(self):
    """FlashBuildEsptool capability to flash bin image."""
    return self.lazy_init(
        flash_build_esptool.FlashBuildEsptool,
        device_name=self.name,
        chip_type=_CHIP_TYPE,
        serial_port=self.communication_address,
        switchboard=self.switchboard,
        wait_for_bootup_complete_fn=self.wait_for_bootup_complete,
        reset_endpoints_fn=self.matter_endpoints.reset,
        boot_up_time=_DEFAULT_BOOTUP_TIMEOUT_SECONDS,
        baud=_BAUDRATE,
        flash_mode=_FLASH_MODE,
        flash_freq=_FLASH_FREQ)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for an PwRPC (Pigweed RPC) device common capability."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class PwRPCCommonBase(capability_base.CapabilityBase):
  """Pigweed RPC common capability for devices communicating over PwRPC."""

  @property
  @abc.abstractmethod
  def vendor_id(self) -> int:
    """The vendor id of the device."""

  @property
  @abc.abstractmethod
  def product_id(self) -> int:
    """The product id of the device."""

  @property
  @abc.abstractmethod
  def software_version(self) -> int:
    """The software version of the device."""

  @abc.abstractmethod
  def reboot(self) -> None:
    """Reboots the device."""

  @abc.abstractmethod
  def factory_reset(self) -> None:
    """Factory resets the device."""

  @abc.abstractmethod
  def ota(self) -> None:
    """Triggers OTA to the device."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter spec enum module."""
import enum
from gazoo_device.protos import attributes_service_pb2

# The Matter cluster enums definitions: (only enums used in GDM are defined)


class ColorControlCluster(enum.IntEnum):
  """Color control cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_COLOR_CONTROL_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_CURRENT_HUE = 0x0000
  ATTRIBUTE_CURRENT_SATURATION = 0x0001


class DoorLockCluster(enum.IntEnum):
  """Door lock cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_DOOR_LOCK_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_LOCK_STATE = 0x0000


class LevelControlCluster(enum.IntEnum):
  """Level control cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_LEVEL_CONTROL_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_CURRENT_LEVEL = 0x0000


class OccupancySensingCluster(enum.IntEnum):
  """Occupacny sensing cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_OCCUPANCY_SENSING_CLUSTER_ID

  # Attributes ID
  ATTRIBUTE_OCCUPANCY = 0x0000
  ATTRIBUTE_OCCUPANCY_SENSOR_TYPE = 0x0001
  ATTRIBUTE_OCCUPANCY_SENSOR_TYPE_BITMAP = 0x0002


class OnOffCluster(enum.IntEnum):
  """OnOff cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_ON_OFF_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_ON_OFF = 0x0000


class TemperatureMeasurementCluster(enum.IntEnum):
  """Temperature Measurement cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_TEMP_MEASUREMENT_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_MEASURED_VALUE = 0
  ATTRIBUTE_MIN_MEASURED_VALUE = 1
  ATTRIBUTE_MAX_MEASURED_VALUE = 2


# Lock state attribute enums
class LockState(enum.IntEnum):
  """Lock state attribute values.
  """
  NOT_FULLY_LOCKED = 0
  LOCKED = 1
  UNLOCKED = 2


# Occupancy Sensor type enums.
class OccupancySensorType(enum.IntEnum):
  """Occupancy Sensor type.
  """
  PIR = 0
  ULTRASONIC = 1
  PIR_AND_ULTRASONIC = 2
  PHYSICAL_CONTACT = 3



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the Pigweed RPC device common capability."""
import time
from typing import Any, Callable, List, Optional
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_common_base
from gazoo_device.protos import device_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport
from gazoo_device.utility import retry


logger = gdm_logger.get_logger()
_POLL_INTERVAL_SEC = 0.5  # seconds
_DEFAULT_BOOTUP_TIMEOUT = 30  # seconds


class PwRPCCommonDefault(pwrpc_common_base.PwRPCCommonBase):
  """Pigweed RPC common capability for devices communicating over PwRPC."""

  def __init__(self,
               device_name: str,
               switchboard_call: Callable[..., Any],
               rpc_timeout_s: int):
    """Create an instance of the PwRPCCommonDefault capability.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The switchboard.call method which calls to the endpoint.
      rpc_timeout_s: Timeout (s) for RPC call.
    """
    super().__init__(device_name=device_name)
    self._switchboard_call = switchboard_call
    self._rpc_timeout_s = rpc_timeout_s

  @decorators.DynamicProperty
  def vendor_id(self) -> int:
    """The vendor id of the device."""
    return self.get_device_info().vendor_id

  @decorators.DynamicProperty
  def product_id(self) -> int:
    """The product id of the device."""
    return self.get_device_info().product_id

  @decorators.DynamicProperty
  def software_version(self) -> int:
    """The software version of the device."""
    return self.get_device_info().software_version

  @decorators.DynamicProperty
  def qr_code(self) -> str:
    """The pairing QR code of the device."""
    return self.pairing_info.qr_code

  @decorators.DynamicProperty
  def qr_code_url(self) -> str:
    """The pairing QR code URL of the device."""
    return self.pairing_info.qr_code_url

  @decorators.DynamicProperty
  def pairing_state(self) -> bool:
    """The pairing state of the device.

    The device is paired if it has at least one fabric ID and the fabric
    contains a node ID.

    Returns:
      The pairing state.
    """
    return bool(self.fabric_info) and hasattr(self.fabric_info[0], "node_id")

  @decorators.DynamicProperty
  def pairing_info(self) -> device_service_pb2.PairingInfo:
    """The pairing information of the device."""
    return self.get_device_info().pairing_info

  @decorators.DynamicProperty
  def fabric_info(self) -> List[device_service_pb2.FabricInfo]:
    """The list of fabric information of the device."""
    repeated_fabric_info = self.get_device_state().fabric_info
    return list(repeated_fabric_info)

  @decorators.CapabilityLogDecorator(logger)
  def reboot(self,
             verify: bool = True,
             bootup_timeout_s: int = _DEFAULT_BOOTUP_TIMEOUT) -> None:
    """Reboots the device.

    Args:
      verify: If true, waits for device bootup completes before returning.
      bootup_timeout_s: Timeout (s) to wait for the device to boot up.
    """
    self._trigger_device_action(action="Reboot")
    if verify:
      self.wait_for_bootup_complete(bootup_timeout_s)

  @decorators.CapabilityLogDecorator(logger)
  def factory_reset(self,
                    verify: bool = True,
                    bootup_timeout_s: int = _DEFAULT_BOOTUP_TIMEOUT) -> None:
    """Factory resets the device.

    Args:
      verify: If true, waits for device bootup completes before returning.
      bootup_timeout_s: Timeout (s) to wait for the device to boot up.
    """
    self._trigger_device_action(action="FactoryReset")
    if verify:
      self.wait_for_bootup_complete(bootup_timeout_s)

  @decorators.CapabilityLogDecorator(logger)
  def ota(self) -> None:
    """Triggers OTA to the device."""
    self._trigger_device_action(action="TriggerOta")

  @decorators.CapabilityLogDecorator(logger)
  def wait_for_bootup_complete(self, bootup_timeout: int) -> None:
    """Waits for device to boot up.

    Args:
      bootup_timeout: Max time to wait for bootup to complete (in seconds).

    Raises:
      DeviceError: If device did not boot up successfully in given timeout.
    """
    try:
      start_time = time.time()
      retry.retry(
          func=self.get_device_info,
          timeout=bootup_timeout,
          interval=_POLL_INTERVAL_SEC,
          reraise=False)
      logger.info(
          f"{self._device_name} booted up in {time.time() - start_time}s.")
    except errors.CommunicationTimeoutError as bootup_error:
      raise errors.DeviceError(
          f"{self._device_name} failed to boot up within {bootup_timeout}s."
      ) from bootup_error

  @decorators.CapabilityLogDecorator(logger)
  def get_device_info(self) -> device_service_pb2.DeviceInfo:
    """Returns device static information.

    Raises:
      DeviceError: The ack status is not true.
    """
    payload_in_bytes = self._trigger_device_action(action="GetDeviceInfo")
    payload = device_service_pb2.DeviceInfo.FromString(payload_in_bytes)
    return payload

  @decorators.CapabilityLogDecorator(logger)
  def get_device_state(self) -> device_service_pb2.DeviceState:
    """Returns device state information.

    Raises:
      DeviceError: The ack status is not true.
    """
    payload_in_bytes = self._trigger_device_action(action="GetDeviceState")
    payload = device_service_pb2.DeviceState.FromString(payload_in_bytes)
    return payload

  @decorators.CapabilityLogDecorator(logger)
  def set_pairing_info(
      self, code: Optional[int] = None, discriminator: Optional[int] = None
    ) -> None:
    """Sets the pairing info of the device.

    Args:
      code: New pairing code to set.
      discriminator: New discriminator to set.
    """
    new_code = self.pairing_info.code if code is None else code
    new_discriminator = (
        self.pairing_info.discriminator if discriminator is None
        else discriminator)
    self._trigger_device_action(action="SetPairingInfo",
                                code=new_code,
                                discriminator=new_discriminator)

  def _trigger_device_action(self, action: str, **kwargs: Any) -> bytes:
    """Triggers specific device action.

    Args:
      action: Device actions including reboot, factory-reset,
        OTA and setting/getting device state and pairing state.
      **kwargs: Arguments for the device action.

    Returns:
      The payload in bytes format.

    Raises:
      DeviceError: The ack status is not true.
    """
    device_rpc_kwargs = {"pw_rpc_timeout_s": self._rpc_timeout_s}
    device_rpc_kwargs.update(kwargs)
    ack, payload_in_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=("Device", action),
        method_kwargs=device_rpc_kwargs)

    if not ack:
      raise errors.DeviceError(f"{self._device_name} triggering {action} failed"
                               ": The action did not succeed")

    return payload_in_bytes



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Build the gazoo_device Python package."""
import os
import re
from typing import List

import setuptools

_CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
_README_FILE = 'README.md'
_REQUIREMENTS_FILE = 'requirements.txt'
_SOURCE_CODE_DIR_NAME = 'gazoo_device'
_VERSION_FILE = '_version.py'


def _get_readme() -> str:
  """Returns contents of README.md."""
  readme_file = os.path.join(_CURRENT_DIR, _README_FILE)
  with open(readme_file) as open_file:
    return open_file.read()


def _get_requirements() -> List[str]:
  """Returns package requirements from requirements.txt."""
  requirements_file = os.path.join(_CURRENT_DIR, _REQUIREMENTS_FILE)
  with open(requirements_file) as open_file:
    requirements = open_file.read()
    if requirements:
      return requirements.splitlines()
  raise RuntimeError('Unable to find package requirements in {}'.format(
      requirements_file))


def _get_version() -> str:
  """Returns version from version.py."""
  version_file = os.path.join(
      _CURRENT_DIR, _SOURCE_CODE_DIR_NAME, _VERSION_FILE)
  with open(version_file) as open_file:
    contents = open_file.read()
    version_match = re.search(r'^version = "(.*)"', contents, re.M)
    if version_match:
      return version_match.group(1)
  raise RuntimeError(f'Unable to find "version = <version>" in {version_file}')


setuptools.setup(
    name='gazoo_device',
    version=_get_version(),
    description='Gazoo Device Manager',
    long_description=_get_readme(),
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing',
    ],

    # project's homepage
    url='https://github.com/google/gazoo-device',

    # author details
    author='Google LLC',
    author_email='gdm-authors@google.com',
    license='Apache 2.0',

    # define list of packages included in distribution
    packages=setuptools.find_packages(exclude=['gazoo_device.tests*']),
    package_dir={'gazoo_device': _SOURCE_CODE_DIR_NAME},
    package_data={
        'gazoo_device': [],
    },

    # runtime dependencies that are installed by pip during install
    install_requires=_get_requirements(),
    entry_points={
        'console_scripts': [
            'gdm=gazoo_device.gdm_cli:main'
        ],
    })



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Raspberry Pi Matter Controller device class."""

from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.auxiliary_devices import raspberry_pi
from gazoo_device.capabilities import matter_controller_chip_tool

logger = gdm_logger.get_logger()


class RaspberryPiMatterController(raspberry_pi.RaspberryPi):
  """Base Class for RaspberryPiMatterController Devices."""
  DETECT_MATCH_CRITERIA = {
      detect_criteria.SshQuery.IS_RPI: True,
      detect_criteria.SshQuery.IS_CHIP_TOOL_PRESENT: True,
  }
  DEVICE_TYPE = "rpi_matter_controller"
  _OWNER_EMAIL = "gdm-authors@google.com"

  @decorators.CapabilityDecorator(
      matter_controller_chip_tool.MatterControllerChipTool)
  def matter_controller(
      self) -> matter_controller_chip_tool.MatterControllerChipTool:
    """Matter controller capability to send chip-tool commands to the device."""
    return self.lazy_init(
        matter_controller_chip_tool.MatterControllerChipTool,
        device_name=self.name,
        regex_shell_fn=self.shell_with_regex,
        shell_fn=self.shell,
        send_file_to_device=self.file_transfer.send_file_to_device)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC implementation of the Matter Door Lock cluster capability.
"""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import door_lock_base
from gazoo_device.protos import attributes_service_pb2

logger = gdm_logger.get_logger()
DoorLockCluster = matter_enums.DoorLockCluster
BOOLEAN_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_BOOLEAN_ATTRIBUTE_TYPE


class DoorLockClusterPwRpc(door_lock_base.DoorLockClusterBase):
  """Matter Door Lock cluster capability."""

  @decorators.CapabilityLogDecorator(logger)
  def lock_door(self, verify: bool = True) -> None:
    """The Lock Door command to lock the Matter device.

    Args:
      verify: If true, verifies the state configurations before returning.
    """
    self._lock_command(lock=True, verify=verify)

  @decorators.CapabilityLogDecorator(logger)
  def unlock_door(self, verify: bool = True) -> None:
    """The Unlock Door command to unlock the Matter device.

    Args:
      verify: If true, verifies the state configurations before returning.
    """
    self._lock_command(lock=False, verify=verify)

  @decorators.DynamicProperty
  def lock_state(self) -> int:
    """The LockState attribute.

    Returns:
      The attribute value of the current locked state.
    """
    locked_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=DoorLockCluster.ID,
        attribute_id=DoorLockCluster.ATTRIBUTE_LOCK_STATE,
        attribute_type=BOOLEAN_ATTRIBUTE_TYPE)
    return (matter_enums.LockState.LOCKED if locked_data.data_bool
            else matter_enums.LockState.UNLOCKED)

  def _lock_command(self, lock: bool, verify: bool = True) -> None:
    """Locks or unlocks the device.

    Args:
      lock: Locks the device if true, unlocks the device if false.
      verify: If true, verifies the lock configurations before returning.

    Raises:
      DeviceError: Ack value is false or the device does not transition to the
      appropriate state.
    """
    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=DoorLockCluster.ID,
        attribute_id=DoorLockCluster.ATTRIBUTE_LOCK_STATE,
        attribute_type=BOOLEAN_ATTRIBUTE_TYPE,
        data_bool=lock)

    if verify:
      expected_state = (matter_enums.LockState.LOCKED if lock else
                        matter_enums.LockState.UNLOCKED)
      if expected_state != self.lock_state:  # pylint: disable=comparison-with-callable
        raise errors.DeviceError(
            f"Device {self._device_name} lock state attribute did not change "
            f"from {self.lock_state}.")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter endpoint and cluster modules."""

from gazoo_device.capabilities.matter_clusters import color_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import door_lock_pw_rpc
from gazoo_device.capabilities.matter_clusters import level_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import occupancy_pw_rpc
from gazoo_device.capabilities.matter_clusters import on_off_pw_rpc
from gazoo_device.capabilities.matter_clusters import temperature_measurement_pw_rpc
from gazoo_device.capabilities.matter_endpoints import color_temperature_light
from gazoo_device.capabilities.matter_endpoints import dimmable_light
from gazoo_device.capabilities.matter_endpoints import door_lock
from gazoo_device.capabilities.matter_endpoints import on_off_light
from gazoo_device.capabilities.matter_endpoints import temperature_sensor
import immutabledict


SUPPORTED_ENDPOINTS = (
    color_temperature_light.ColorTemperatureLightEndpoint,
    dimmable_light.DimmableLightEndpoint,
    door_lock.DoorLockEndpoint,
    on_off_light.OnOffLightEndpoint,
    temperature_sensor.TemperatureSensorEndpoint)

SUPPORTED_CLUSTERS = (
    color_control_pw_rpc.ColorControlClusterPwRpc,
    door_lock_pw_rpc.DoorLockClusterPwRpc,
    level_control_pw_rpc.LevelControlClusterPwRpc,
    occupancy_pw_rpc.OccupancyClusterPwRpc,
    on_off_pw_rpc.OnOffClusterPwRpc,
    temperature_measurement_pw_rpc.TemperatureMeasurementClusterPwRpc)

MATTER_DEVICE_TYPE_ID_TO_CLASS = immutabledict.immutabledict({
    endpoint_class.DEVICE_TYPE_ID: endpoint_class
    for endpoint_class in SUPPORTED_ENDPOINTS
})

CLUSTER_ID_TO_CLASS = immutabledict.immutabledict({
    cluster_class.CLUSTER_ID: cluster_class
    for cluster_class in SUPPORTED_CLUSTERS
})



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gets information on all usb connected devices.

Returns a dictionary mapping a persistent address (adb or serial port) to usb
info for that address.

Note: usb_info_linux primarily relies on pyudev for full information.
It processes just adb devices and those that show up in /dev/serial/by-id
"""
import re
import sys

from gazoo_device.utility import usb_config
import pyudev


PERSISTENT_SERIAL_PATH_FOLDER = '/dev/serial/by-id'
MOUNTED_DISK_FOLDER = '/dev/disk/by-id'
ADB_SERIAL_REGEX = r'\/dev\/disk\/by-id\/usb-Linux_File-CD_Gadget_([\da-z]+)-0:0'
HUB_IDENTIFIER_REGEX = r'\d-\d(?:\.\d?)+/(\d-(?:\d\.)+)\d\.\d:\d\.\d'
USB3_HUB_IDENTIFIER_REGEX = r'\d-\d(?:\.\d?)+/(\d-(?:\d\.)+)\d:\d\.\d'
IDENTIFIER_REGEX_TEMPLATE = r'\d-\d(?:\.\d?)+/{hub}(\d\.\d)(?:\.\d)*:\d\.\d'
ANDROID_DEVICE_PRODUCT_NAME = 'File-CD Gadget'

# These vendor's mounted drives show up as multiple usb connections.
# We use the one with the symlink.
REPEAT_VENDORS = [
    'FTDI',
    'cambrionix',
    'Linux',
    'Android',
]


def get_address_to_usb_info_dict():
  """Returns a dict of address to the usb information.

  Returns:
     dict: dictionary of address to UsbInfo instance for each usb device
     attached.
  """
  udev_devices = get_pyudev_list_of_devices()
  relevant_devices = [
      device for device in udev_devices
      if device.properties.get('ID_BUS') == 'usb'
  ]
  address_to_usb_info_dict = {}
  location_dict = {}
  model_dict = {}
  for device in relevant_devices:
    address, entry, dev_path, model = _process_udev_device(device)
    if address:  # remove devices that don't have good addresses
      address_to_usb_info_dict[address] = entry
      if dev_path:
        location_dict[address] = dev_path
      if model:
        model_dict[address] = model
  disk_info = _get_disk_info()
  _add_disk_info(address_to_usb_info_dict, disk_info)
  _add_cambrionix_ports(address_to_usb_info_dict, location_dict, model_dict)
  return address_to_usb_info_dict


def get_pyudev_list_of_devices(subsystem=None, devtype=None):
  """Wrapper around pyudev.list_devices.

  Args:
    subsystem (str): maps to pyudev.Context().list_devices arg.
    devtype (str): maps to pyudev.Context().list_devices arg DEVTYPE.

  Returns:
    list: a list of pyudev devices.

  Raises:
    RuntimeError: if platform is not linux.
  """
  args = {}
  if subsystem:
    args['subsystem'] = subsystem
  if devtype:
    args['DEVTYPE'] = devtype
  if sys.platform == 'darwin':
    raise RuntimeError('This method is only supported on linux.')
  context = pyudev.Context()
  return list(context.list_devices(**args))


def _add_cambrionix_ports(address_to_usb_info_dict, location_dict, model_dict):
  """Add cambrionix port number and parent address to each cambrionix child."""
  for hub_address, model in model_dict.items():
    for child_address in address_to_usb_info_dict[hub_address].child_addresses:
      if child_address in address_to_usb_info_dict:
        child_entry = address_to_usb_info_dict[child_address]
        child_entry.usb_hub_address = hub_address
        child_entry.usb_hub_port = _get_cambrionix_port_number(
            hub_address, model, child_address, location_dict)


def _add_disk_info(address_to_usb_info_dict, disk_info):
  """Add disk path to the entries that have them."""
  for entry in address_to_usb_info_dict.values():
    serial = entry.serial_number
    if serial in disk_info:
      entry.disk = disk_info[serial]


def _get_address(udev_device):
  """Gets the serial port symlink under '/dev/serial/by-id', mounted drives, or adb serial."""
  devlinks_str = udev_device.properties.get('DEVLINKS', '')
  paths = devlinks_str.split()
  if not devlinks_str:  # Get mounted drives without corresponding symlinks
    if udev_device.properties.get('ID_VENDOR') not in REPEAT_VENDORS:
      return udev_device.properties.get('DEVNAME')

  for path in paths:
    if PERSISTENT_SERIAL_PATH_FOLDER in path:
      return path
    if MOUNTED_DISK_FOLDER in path:
      match = re.match(ADB_SERIAL_REGEX, path)
      if match:
        return match.group(1)
      return path

  return ''


def _get_cambrionix_model(udev_device):
  """Gets the model of the cambrionix."""
  parent = udev_device.parent.parent.parent.parent
  if udev_device.properties.get('ID_MODEL') in usb_config.CAMBRIONIX_PORT_MAP:
    return udev_device.properties.get('ID_MODEL')
  else:
    return parent.properties.get('ID_MODEL')


def _get_cambrionix_port_number(hub_address,
                                hub_model,
                                child_address,
                                location_dict):
  """Get port from dev path.

  Args:
    hub_address (str): Address for the Cambrionix hub.
    hub_model (str): Model for the Cambrionix hub.
    child_address (str): Address for a child device connected to the hub.
    location_dict (dict): Dictionary to look up the dev_path by address.

  Returns:
    int: Port number or None if not found.

  Note:
    The port number is determined by first finding the Cambionix (or parent)
    identifier. The parent identifier can vary in length, but is found using
    the hub dev path by looking between the '/' and the last one or two digits
    (depending on hub model) before the ':' ("/<parent_identifier>n.n:n.n").
    For children, the two digits following the parent identifier in their
    dev path are used to look up the Cambrionix port.

    Example:
      hub dev path:
        "/devices/pci0000:00/0000:00:14.0/usb1/1-4/1-4.1/1-4.1.2/1-4.1.2.4/"
        "1-4.1.2.4.1/1-4.1.2.4.1:1.0/ttyUSB4/tty/ttyUSB4"
      From the hub dev path "1-4.1.2.4.1:1.0" is used to find that the parent
      identifier is "1-4.1.2".
      child dev path:
        "/devices/pci0000:00/0000:00:14.0/usb1/1-4/1-4.1/1-4.1.2/1-4.1.2.3/"
        "1-4.1.2.3.4/1-4.1.2.3.4:1.2/ttyUSB2/tty/ttyUSB2"
          "1-4.1.2.3.4:1.2" is the relevant number.
          "1-4.1.2" is the parent identifier.
          "3.4" is the child identifier used to look up the port number.
          "1.2" indicates the FTDI interface.
      child dev path (hub device):
        "/devices/pci0000:00/0000:00:14.0/usb1/1-4/1-4.1/1-4.1.2/1-4.1.2.4/"
        "1-4.1.2.4.2/1-4.1.2.4.2.1/1-4.1.2.4.2.1:1.0/tty/ttyACM1"
          "1-4.1.2.4.2.1:1.0" is the relevant number.
          "1-4.1.2" is the parent identifier.
          "4.2" is the child identifier used to look up the port number.
          "1:" is the hub device identifier
          "1.0" indicates the FTDI interface.
  """
  if 'USB3' in hub_model:
    hub_id_regex = USB3_HUB_IDENTIFIER_REGEX
  else:
    hub_id_regex = HUB_IDENTIFIER_REGEX
  hub_match = re.search(hub_id_regex, location_dict[hub_address])
  if hub_match:
    parent_identifier = hub_match.group(1).replace('.', r'\.')
    identifier_regex = IDENTIFIER_REGEX_TEMPLATE.format(hub=parent_identifier)
    match = re.search(identifier_regex, location_dict[child_address])
    if match:
      port_mapping = usb_config.CAMBRIONIX_PORT_MAP[hub_model]
      return port_mapping.get(match.group(1))
  return None


def _get_child_addresses(udev_device):
  """Get the addresses for the relevant children of the udev device."""
  all_children = _get_cambrionix_udev_children(udev_device)
  relevant_children = [
      device for device in all_children
      if device.properties.get('DEVLINKS') and _get_address(device)
  ]
  if udev_device in relevant_children:
    relevant_children.remove(udev_device)
  grand_children = []
  for child_device in relevant_children:
    # Remove children of other cambrionixes
    if _get_product_name(child_device) in usb_config.CAMBRIONIX_NAMES:
      grand_children += _get_cambrionix_udev_children(child_device)
      grand_children.remove(child_device)
  return [
      _get_address(child)
      for child in relevant_children
      if child not in grand_children
  ]


def _get_disk_info():
  """Returns a dictionary of disk paths to serial numbers."""
  devices = get_pyudev_list_of_devices(subsystem='block')
  mounted_info = {
      device.properties.get('ID_SERIAL_SHORT', ''):
          device.properties.get('DEVNAME')
      for device in devices
  }
  return mounted_info


def _get_product_name(udev_device):
  """Gets the product name in a system agnostic way."""
  model = udev_device.properties.get('ID_MODEL', '').replace('_', ' ')
  if model == ANDROID_DEVICE_PRODUCT_NAME:  # Actually an Android device.
    model = 'Android'
  return model


def _process_udev_device(udev_device):
  """Convert pyudev dict to the system agnostic form."""
  model = None
  dev_path = udev_device.properties.get('DEVPATH', '')

  entry = usb_config.UsbInfo()
  address = _get_address(udev_device)
  if not address:
    return '', {}, None, None
  entry.address = address
  entry.product_id = udev_device.properties.get('ID_MODEL_ID', '')
  entry.vendor_id = udev_device.properties.get('ID_VENDOR_ID', '')
  entry.ftdi_interface = int(
      udev_device.properties.get('ID_USB_INTERFACE_NUM', 0))
  entry.manufacturer = udev_device.properties.get('ID_VENDOR', '')
  entry.serial_number = udev_device.properties.get('ID_SERIAL_SHORT', '')
  entry.product_name = _get_product_name(udev_device)
  if entry.product_name in usb_config.CAMBRIONIX_NAMES:
    entry.child_addresses = _get_child_addresses(udev_device)
    model = _get_cambrionix_model(udev_device)
  return address, entry, dev_path, model


def _get_cambrionix_udev_children(udev_device):
  """Gets all the udev devices representing a cambrionix's children."""
  return list(udev_device.parent.parent.parent.parent.parent.children)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handles registration of device extension packages.

Registered extensions are stored in gazoo_device.extensions.
"""
import collections
import copy
import importlib
import inspect
import itertools
import json
import logging
import os.path
import types
from typing import (
    Any, Callable, Collection, List, Mapping, Optional, Tuple, Type, Union)

from gazoo_device import config
from gazoo_device import data_types
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import extensions
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.base_classes import gazoo_device_base
from gazoo_device.capabilities.interfaces import capability_base
from gazoo_device.capabilities.interfaces import switchboard_base
from gazoo_device.switchboard import communication_types
from gazoo_device.utility import common_utils
from gazoo_device.utility import conformance_utils
import immutabledict

# Temporary exceptions to the conformance checks
OVERRIDDEN_SIGNATURE_DEVICE_METHODS = (
)
PUBLIC_METHOD_EXCEPTIONS = (
)

_EXTENSION_DICT_DEFAULTS = immutabledict.immutabledict({
    "primary_devices": [],
    "auxiliary_devices": [],
    "virtual_devices": [],
    "communication_types": [],
    "detect_criteria": {},
    "capability_interfaces": [],
    "capability_flavors": [],
    "keys": [],
    "manager_cli_mixin": None,
})
_EXPECTED_QUERY_ARGS = ("address", "detect_logger", "create_switchboard_func")
_MISMATCHING_SIGNATURE_TEMPLATE = (
    "Method {!r}, child signature {}, inherited signature(s) {}.")
_DetectQueryType = Callable[
    [str, logging.Logger, Callable[..., switchboard_base.SwitchboardBase]],
    Union[str, bool]
]
_DetectQueryMapping = Mapping[detect_criteria.QueryEnum, _DetectQueryType]
_VIRTUAL_ENV_PIP_PATH = os.path.join(config.VIRTUAL_ENV_DIRECTORY, "bin", "pip")

_AuxiliaryDeviceBase = auxiliary_device.AuxiliaryDevice
_PrimaryDeviceBase = gazoo_device_base.GazooDeviceBase
# Virtual devices currently do not have a dedicated base class.
_VirtualDeviceBase = gazoo_device_base.GazooDeviceBase
_CommunicationTypeBase = communication_types.CommunicationType
_CapabilityBase = capability_base.CapabilityBase
_DeviceClassType = Union[_AuxiliaryDeviceBase, _PrimaryDeviceBase,
                         _VirtualDeviceBase]

logger = gdm_logger.get_logger()


def register(package: types.ModuleType) -> None:
  """Registers the given extension package with GDM architecture.

  The provided module must define:
  * a "download_key" function;
  * a __version__ attribute (must be a string);
  * an "export_extensions" function.

  The download_key function must have the following signature:
  download_key(key_info: data_types.KeyInfo, local_key_path: str) -> None.
  It will be called if one of the SSH keys registered by the package is missing
  on the host. The function must obtain the key and write it to local_key_path.
  key_info is guaranteed to be one of the keys registered by the package.
  Directory containing local_key_path is guaranteed to exist. Key permissions
  for SSH keys will be set by GDM automatically.

  The export_extensions function must have the following signature:
  export_extensions() -> Dict[str, Any].
  The return dictionary must contain all extensions the package wishes to
  register with the GDM architecture. The dictionary format must be:
  {
      "primary_devices": [  # list of GazooDeviceBase subclasses
          <PrimaryDeviceClass1>, <...>
      ],
      "auxiliary_devices": [  # list of AuxiliaryDevice subclasses
          <AuxiliaryDeviceClass1>, <...>
      ],
      "virtual_devices": [  # list of GazooDeviceBase subclasses
          <VirtualDeviceClass1>, <...>
      ],
      "communication_types": [  # list of CommunicationType subclasses
          <CommunicationTypeClass1>, <...>
      ],
      "detect_criteria": {  # CommTypeName -> {QueryEnum -> Callable} entries
          "CommunicationTypeName1": {
              <DetectionQueryName1>: <detection_query_1_function>,
              <DetectionQueryName2>: <detection_query_2_function>,
              <...>
          },
          "CommunicationTypeName2": {
              <...>
          }
      },
      "capability_interfaces": [  # CapabilityBase subclasses (usually abstract)
          <Capability1Base>, <Capability2Base>, <...>
      ],
      "capability_flavors": [  # concrete CapabilityBase subclasses
          <Capability1Foo>, <Capability1Bar>, <Capability2Default>, <...>
      ],
      "keys": [  # data_types.KeyInfo instances
          <KeyInfoSsh1>, <KeyInfoSsh2>, <KeyInfoApiToken1>, <...>
      ],
      # Manager mixins are used to expose package-defined CLI commands.
      "manager_cli_mixin": <class object inheriting from FireManager> or None,
  }

  Args:
    package: Extension package to register (typically its __init__ module).

  Raises:
    PackageRegistrationError: The extensions provided by
      package.export_extensions() are invalid, or package.__version__ is not
      defined.
  """
  package_name = package.__name__.split(".")[-1]
  _validate_extension_package(package, package_name)

  new_extensions = dict(copy.deepcopy(_EXTENSION_DICT_DEFAULTS))
  new_extensions.update(package.export_extensions())
  extensions_backup = _copy_extensions()

  try:
    _register(new_extensions, package_name,
              package.__version__,  # pytype: disable=attribute-error
              package.__name__,
              package.download_key)
  except errors.PackageRegistrationError:
    # Registration failed: revert all changes to the extensions to avoid
    # leaving them in an inconsistent state.
    _restore_extensions(extensions_backup)
    raise


def import_and_register(package_name: str,
                        include_cli_instructions: bool = False) -> bool:
  """Attempts to import and register the extension package.

  Args:
    package_name: Name of the package to import. For example,
      "foo_extension_package" or "my_package.bar_devices".
    include_cli_instructions: Whether to include CLI-specific instructions to
      resolve the error.

  Returns:
    True if operation succeeded, False otherwise.
  """
  try:
    package = importlib.import_module(package_name)
  except ImportError:
    logger.warning(
        f"Import of GDM extension package {package_name!r} failed. "
        "GDM will not be able to use the package. Error:",
        exc_info=True)
    if include_cli_instructions:
      logger.warning(
          "\nInstall the package in the GDM CLI virtual environment via "
          f"`{_VIRTUAL_ENV_PIP_PATH} install <package>` or unregister the "
          f"package from GDM CLI via `gdm unregister {package_name}`\n")
    return False
  try:
    register(package)
  except errors.PackageRegistrationError:
    logger.warning(
        f"Registration of GDM extension package {package_name!r} failed. "
        "GDM will not be able to use the package. Error:",
        exc_info=True)
    if include_cli_instructions:
      logger.warning(
          "\nUpdate the package and install the new version in the GDM CLI "
          "virtual environment via "
          f"`{_VIRTUAL_ENV_PIP_PATH} install <package>` or unregister the "
          f"package from GDM CLI via `gdm unregister {package_name}`\n")
    return False
  return True


def get_cli_extension_packages() -> List[str]:
  """Returns names of extension packages registered with the CLI."""
  if os.path.exists(config.DEFAULT_GDM_CONFIG_FILE):
    with open(config.DEFAULT_GDM_CONFIG_FILE) as gdm_config_file:
      gdm_config = json.load(gdm_config_file)
    return gdm_config.get("cli_extension_packages", [])
  return []


def import_and_register_cli_extension_packages() -> None:
  """Attempts to import and register all packages registered with the CLI."""
  cli_extension_packages = get_cli_extension_packages()
  for package_name in cli_extension_packages:
    import_and_register(package_name, include_cli_instructions=True)


def _validate_extension_package(package: types.ModuleType,
                                package_name: str) -> None:
  """Checks that the extension package is valid.

  Args:
    package: Extension package to check.
    package_name: Name of the extension package.

  Raises:
    PackageRegistrationError: Package has already been registered, or a required
      attribute is missing or invalid.
  """
  if package_name in extensions.package_info:
    raise errors.PackageRegistrationError(
        f"Package {package_name!r} has already been registered.",
        package_name=package_name)

  package_version = getattr(package, "__version__", None)
  if not package_version or not isinstance(package_version, str):
    raise errors.PackageRegistrationError(
        f"Expected __version__ to be a string, found {package_version}.",
        package_name=package_name)

  missing_functions = [
      func for func in ("download_key", "export_extensions")
      if not hasattr(package, func) or not callable(getattr(package, func))
  ]
  if missing_functions:
    raise errors.PackageRegistrationError(
        f"Package must define functions {missing_functions}.",
        package_name=package_name)


def _register(
    new_extensions: Mapping[str, Any],
    package_name: str,
    package_version: str,
    package_import_path: str,
    download_key: Callable[[data_types.KeyInfo, str], None]) -> None:
  """Registers the given extensions with GDM architecture.

  Args:
    new_extensions: Extensions to register with GDM architecture. See the
      register() docstring for format details.
    package_name: Name of the package being registered.
    package_version: Version of the package being registered.
    package_import_path: Import path of the package.
    download_key: download_key function of the package.

  Raises:
    PackageRegistrationError: The provided extensions are invalid.
  """
  _validate_capability_interfaces(
      ext_capability_interfaces=new_extensions["capability_interfaces"],
      package_name=package_name)
  new_capability_interfaces = {
      common_utils.generate_name(interface): interface
      for interface in new_extensions["capability_interfaces"]
  }
  # Capability interfaces have to be registered before validating flavors.
  extensions.capability_interfaces.update(new_capability_interfaces)
  new_capabilities = {
      interface.get_capability_name(): if_name
      for if_name, interface in new_capability_interfaces.items()
  }
  extensions.capabilities.update(new_capabilities)

  _validate_capability_flavors(
      ext_capability_flavors=new_extensions["capability_flavors"],
      package_name=package_name)
  new_capability_flavors = {
      common_utils.generate_name(flavor): flavor
      for flavor in new_extensions["capability_flavors"]
  }
  # Capability flavors have to be registered before validating device classes.
  extensions.capability_flavors.update(new_capability_flavors)

  _validate_device_classes(
      ext_auxiliary_devices=new_extensions["auxiliary_devices"],
      ext_primary_devices=new_extensions["primary_devices"],
      ext_virtual_devices=new_extensions["virtual_devices"],
      package_name=package_name)
  _validate_comm_type_classes(
      ext_communication_types=new_extensions["communication_types"],
      package_name=package_name)
  _validate_detect_criteria(
      ext_detect_criteria=new_extensions["detect_criteria"],
      ext_communication_types=new_extensions["communication_types"],
      package_name=package_name)
  _validate_keys(new_extensions["keys"], package_name=package_name)
  _validate_manager_cli_mixin(new_extensions["manager_cli_mixin"],
                              package_name=package_name)

  extensions.auxiliary_devices = list(set(
      extensions.auxiliary_devices
      + new_extensions["auxiliary_devices"]))
  extensions.primary_devices = list(set(
      extensions.primary_devices
      + new_extensions["primary_devices"]))
  extensions.virtual_devices = list(set(
      extensions.virtual_devices
      + new_extensions["virtual_devices"]))
  extensions.communication_types.update({
      comm_type.__name__: comm_type
      for comm_type in new_extensions["communication_types"]
  })
  for comm_type_name, query_dict in new_extensions["detect_criteria"].items():
    if comm_type_name not in extensions.detect_criteria:
      extensions.detect_criteria[comm_type_name] = dict(query_dict)
    else:
      extensions.detect_criteria[comm_type_name].update(query_dict)
  extensions.keys.extend(new_extensions["keys"])
  if new_extensions["manager_cli_mixin"]:
    extensions.manager_cli_mixins.append(new_extensions["manager_cli_mixin"])

  extensions.package_info[package_name] = immutabledict.immutabledict({
      "version": package_version,
      "key_download_function": download_key,
      "import_path": package_import_path,
  })


def _validate_device_classes(ext_auxiliary_devices: Collection[Type[Any]],
                             ext_primary_devices: Collection[Type[Any]],
                             ext_virtual_devices: Collection[Type[Any]],
                             package_name: str) -> None:
  """Validates the extension device classes.

  Args:
    ext_auxiliary_devices: Auxiliary device classes to validate.
    ext_primary_devices: Primary device classes to validate.
    ext_virtual_devices: Virtual device classes to validate.
    package_name: Name of the package providing the extension classes.

  Raises:
    PackageRegistrationError: Device classes are invalid.
  """
  _assert_subclasses(ext_auxiliary_devices, _AuxiliaryDeviceBase, package_name,
                     "auxiliary device")
  _assert_subclasses(ext_primary_devices, _PrimaryDeviceBase, package_name,
                     "primary device")
  _assert_subclasses(ext_virtual_devices, _VirtualDeviceBase, package_name,
                     "virtual device")

  new_device_classes = tuple(itertools.chain(
      ext_auxiliary_devices, ext_primary_devices, ext_virtual_devices))
  known_device_classes = (extensions.auxiliary_devices
                          + extensions.primary_devices
                          + extensions.virtual_devices)
  new_device_types = tuple(device_class.DEVICE_TYPE
                           for device_class in new_device_classes)
  _assert_unique(new_device_types,
                 names_description="Device types",
                 classes_description="device classes",
                 package_name=package_name)
  redefined_device_types = set.intersection(
      {device_class.DEVICE_TYPE for device_class in new_device_classes},
      {device_class.DEVICE_TYPE for device_class in known_device_classes})
  if redefined_device_types:
    raise errors.PackageRegistrationError(
        f"Device types {redefined_device_types} are already defined in GDM.",
        package_name=package_name)

  conformance_issues = _get_device_class_conformance_issues(new_device_classes)
  if conformance_issues:
    issue_messages = []
    for cls, issues in conformance_issues:
      issue_message = "".join(f"\n\t{issue}" for issue in issues)
      issue_messages.append(f"{cls}{issue_message}")
    raise errors.PackageRegistrationError(
        "The following device class(es) are incompliant with GDM "
        "architecture:\n{}".format("\n".join(issue_messages)),
        package_name=package_name)


def _get_device_class_conformance_issues(
    device_classes: Collection[Type[_DeviceClassType]]
) -> List[Tuple[Type[_DeviceClassType], List[str]]]:
  """Returns conformance issues identified in the device classes.

  Args:
    device_classes: Device classes to check for conformance with architecture.

  Returns:
    List of (device class, list of conformance issues) tuples.
    If there are no issues for a device class, it will not be present in the
    return list.
  """
  all_issues = []
  for device_class in device_classes:
    issues = []
    no_log_decorator_methods = conformance_utils.get_log_decorator_violators(
        device_class)
    if no_log_decorator_methods:
      issues.append(
          "Public methods without return values must be decorated with "
          "@decorators.LogDecorator(<logger>). Incompliant methods: "
          f"{no_log_decorator_methods}.")
    mismatching_signatures = conformance_utils.get_mismatching_signatures(
        device_class, excluded_methods=OVERRIDDEN_SIGNATURE_DEVICE_METHODS)
    if mismatching_signatures:
      mismatching_signature_messages = [
          _MISMATCHING_SIGNATURE_TEMPLATE.format(method, child_sig, parent_sigs)
          for method, child_sig, parent_sigs in mismatching_signatures]
      issues.append(
          "Methods may not fully override signatures inherited from parents. "
          "Only extending the argument list is allowed. "
          f"Incompliant method signatures: {mismatching_signature_messages}.")
    invalid_public_methods = conformance_utils.get_invalid_public_methods(
        device_class, excluded_methods=PUBLIC_METHOD_EXCEPTIONS)
    if invalid_public_methods:
      issues.append(
          "New public methods are not allowed, except for health checks. "
          "Methods must either be private or, if public, moved into "
          f"capabilities. Incompliant methods: {invalid_public_methods}.")
    uncategorized_properties = conformance_utils.get_uncategorized_properties(
        device_class)
    if uncategorized_properties:
      issues.append(
          "Public properties must be categorized as either @decorators"
          ".DynamicProperty, .PersistentProperty, .OptionalProperty, or "
          ".CapabilityDecorator. Incompliant properties: "
          f"{uncategorized_properties}.")
    unset_class_constants = conformance_utils.get_unset_constants(device_class)
    if unset_class_constants:
      issues.append(f"Class constants {unset_class_constants} are not set.")
    invalid_health_checks = conformance_utils.get_invalid_health_check_names(
        device_class)
    if invalid_health_checks:
      issues.append(
          "Health checks must follow the <check_...> naming convention. "
          f"Incompliant health checks: {invalid_health_checks}.")
    invalid_definitions = conformance_utils.get_invalid_capability_definitions(
        device_class)
    if invalid_definitions:
      issues.append("Capability definition(s) are invalid. {}".format(
          ", ".join(f"Capability {cap!r}: {err}"
                    for cap, err in invalid_definitions)))
    if issues:
      all_issues.append((device_class, issues))

  return all_issues


def _validate_comm_type_classes(
    ext_communication_types: Collection[Type[Any]],
    package_name: str) -> None:
  """Validates the extension communication types.

  Args:
    ext_communication_types: Communication type classes to validate.
    package_name: Name of the package providing the extension classes.

  Raises:
    PackageRegistrationError: Communication type classes are invalid.
  """
  _assert_subclasses(ext_communication_types, _CommunicationTypeBase,
                     package_name, "communication type")
  new_comm_types = {
      comm_type.__name__: comm_type
      for comm_type in ext_communication_types
  }
  _raise_if_redefined(new_classes=new_comm_types,
                      old_classes=extensions.communication_types,
                      classes_description="communication types",
                      package_name=package_name)


def _validate_detect_criteria(
    ext_detect_criteria: Mapping[str, _DetectQueryMapping],
    ext_communication_types: Collection[Type[Any]],
    package_name: str) -> None:
  """Validates the extension detection criteria.

  Args:
    ext_detect_criteria: Detection criteria to validate, where mapping keys are
      communication type names, and values are detection query mappings for each
      communication type.
    ext_communication_types: Communication types exported by the package.
    package_name: Name of the package providing the extensions.

  Raises:
    PackageRegistrationError: Detection criteria are invalid.
  """
  extension_comm_type_names = [comm_type.__name__
                               for comm_type in ext_communication_types]
  for comm_type_name, query_dict in ext_detect_criteria.items():
    for query_name, query in query_dict.items():
      base_error = ("Unable to register query {} for communication type {!r}. "
                    .format(query_name, comm_type_name))
      if not isinstance(query_name, detect_criteria.QueryEnum):
        raise errors.PackageRegistrationError(
            base_error + "Detection query keys must be {} instances.".format(
                detect_criteria.QueryEnum.__name__),
            package_name=package_name)
      if (not callable(query)
          or tuple(inspect.getfullargspec(query).args) != _EXPECTED_QUERY_ARGS):
        extra_error = ("Detection queries must be callable functions which "
                       "accept {} arguments: {}."
                       .format(len(_EXPECTED_QUERY_ARGS), _EXPECTED_QUERY_ARGS))
        raise errors.PackageRegistrationError(base_error + extra_error,
                                              package_name=package_name)

    if comm_type_name not in extensions.detect_criteria:
      if comm_type_name not in extension_comm_type_names:
        raise errors.PackageRegistrationError(
            "Unable to register detection criteria for communication type "
            f"{comm_type_name!r} as it has not been exported by the package.",
            package_name=package_name)
    else:
      redefined_queries = list(
          extensions.detect_criteria[comm_type_name].keys()
          & query_dict.keys())
      if redefined_queries:
        raise errors.PackageRegistrationError(
            f"Detection queries {redefined_queries} for communication type "
            f"{comm_type_name!r} are already defined in GDM.",
            package_name=package_name)


def _validate_capability_interfaces(
    ext_capability_interfaces: Collection[Type[Any]],
    package_name: str) -> None:
  """Validates the extension capability interfaces.

  Args:
    ext_capability_interfaces: Capability interface classes to validate.
    package_name: Name of the package providing the extension classes.

  Raises:
    PackageRegistrationError: Capability interface classes are invalid.
  """
  _assert_subclasses(ext_capability_interfaces, _CapabilityBase, package_name,
                     "capability interface", allow_abstract=True)
  new_capability_interfaces = {
      common_utils.generate_name(interface): interface
      for interface in ext_capability_interfaces
  }
  _assert_unique(new_capability_interfaces.keys(),
                 names_description="Capability interface names",
                 classes_description="capability interfaces",
                 package_name=package_name)
  _raise_if_redefined(new_classes=new_capability_interfaces,
                      old_classes=extensions.capability_interfaces,
                      classes_description="capability interfaces",
                      package_name=package_name)

  new_capability_names_list = []  # Entries are (interface name, interface)
  for interface in ext_capability_interfaces:
    try:
      interface_name = interface.get_capability_name()
    except TypeError:
      # Interface did not override name generation logic and isn't recognized
      # by GDM yet.
      interface_name = capability_base.get_default_capability_name(interface)
    new_capability_names_list.append((interface_name, interface))

  _assert_unique(tuple(name for name, _ in new_capability_names_list),
                 names_description="Capability names",
                 classes_description="capabilities",
                 package_name=package_name)
  new_capability_names = dict(new_capability_names_list)
  existing_capability_names = {
      cap_name: extensions.capability_interfaces[if_name]
      for cap_name, if_name in extensions.capabilities.items()
  }
  _raise_if_redefined(new_classes=new_capability_names,
                      old_classes=existing_capability_names,
                      classes_description="capabilities",
                      package_name=package_name)
  _assert_unique((list(new_capability_interfaces.keys())
                  + list(new_capability_names.keys())),
                 names_description="Capability interface and capability names",
                 classes_description="capability interfaces and capabilities",
                 package_name=package_name)


def _validate_capability_flavors(ext_capability_flavors: Collection[Type[Any]],
                                 package_name: str) -> None:
  """Validates the extension capability flavors.

  Args:
    ext_capability_flavors: Capability flavor classes to validate.
    package_name: Name of the package providing the extension classes.

  Raises:
    PackageRegistrationError: Capability flavor classes are invalid.
  """
  _assert_subclasses(ext_capability_flavors, _CapabilityBase, package_name,
                     "capability flavor")
  new_capability_flavors = {
      common_utils.generate_name(flavor): flavor
      for flavor in ext_capability_flavors
  }
  _assert_unique(new_capability_flavors.keys(),
                 names_description="Capability flavor names",
                 classes_description="capability flavors",
                 package_name=package_name)
  _raise_if_redefined(new_classes=new_capability_flavors,
                      old_classes=extensions.capability_flavors,
                      classes_description="capability flavors",
                      package_name=package_name)

  _assert_unique((list(new_capability_flavors.keys())
                  + list(extensions.capability_interfaces.keys())
                  + list(extensions.capabilities.keys())),
                 names_description="Capability flavor and (capability "
                                   "interface or capability) names",
                 classes_description="capability flavors and (capability "
                                     "interfaces or capabilities)",
                 package_name=package_name)

  conformance_issues = _get_capability_flavor_conformance_issues(
      ext_capability_flavors)
  if conformance_issues:
    issue_messages = []
    for cls, issues in conformance_issues:
      issue_message = "".join(f"\n\t{issue}" for issue in issues)
      issue_messages.append(f"{cls}{issue_message}")
    raise errors.PackageRegistrationError(
        "The following capability class(es) are incompliant with GDM "
        "architecture:\n{}".format("\n".join(issue_messages)),
        package_name=package_name)


def _get_capability_flavor_conformance_issues(
    capability_flavors: Collection[Type[_CapabilityBase]]
) -> List[Tuple[Type[_CapabilityBase], List[str]]]:
  """Returns conformance issues identified in the capability flavor classes.

  Args:
    capability_flavors: Capability flavor classes to check for conformance.

  Returns:
    List of (capability flavor class, list of conformance issues) tuples.
    If there are no issues for a capability flavor class, it will not be present
    in the return list.
  """
  all_issues = []
  for capability_flavor in capability_flavors:
    issues = []
    no_log_decorator_methods = conformance_utils.get_log_decorator_violators(
        capability_flavor)
    if no_log_decorator_methods:
      issues.append(
          "Public methods without return values must be decorated with "
          "@decorators.CapabilityLogDecorator(<logger>). "
          f"Incompliant methods: {no_log_decorator_methods}.")
    mismatching_signatures = conformance_utils.get_mismatching_signatures(
        capability_flavor)
    if mismatching_signatures:
      mismatching_signature_messages = [
          _MISMATCHING_SIGNATURE_TEMPLATE.format(method, child_sig, parent_sigs)
          for method, child_sig, parent_sigs in mismatching_signatures]
      issues.append(
          "Methods may not fully override signatures inherited from parents. "
          "Only extending the argument list is allowed. "
          f"Incompliant method signatures: {mismatching_signature_messages}.")
    uncategorized_properties = conformance_utils.get_uncategorized_properties(
        capability_flavor)
    if uncategorized_properties:
      issues.append(
          "Public properties must be categorized as either @decorators"
          ".DynamicProperty, .PersistentProperty, or .OptionalProperty. "
          f"Incompliant properties: {uncategorized_properties}.")
    if issues:
      all_issues.append((capability_flavor, issues))

  return all_issues


def _validate_keys(keys: Collection[data_types.KeyInfo],
                   package_name: str) -> None:
  """Validates the extension keys.

  Args:
    keys: Keys provided by the extension package.
    package_name: Name of the extension package.

  Raises:
    PackageRegistrationError: Provided keys are invalid.
  """
  if not all(isinstance(key, data_types.KeyInfo) for key in keys):
    raise errors.PackageRegistrationError(
        "Keys must be data_types.KeyInfo instances.",
        package_name=package_name)
  if not all(key.package == package_name for key in keys):
    raise errors.PackageRegistrationError(
        "KeyInfo.package attribute must match the name of the package "
        f"({package_name!r}).", package_name=package_name)


def _validate_manager_cli_mixin(manager_cli_mixin: Optional[Type[Any]],
                                package_name: str) -> None:
  """Validates the provided FireManager mixin.

  Args:
    manager_cli_mixin: None or a class object inheriting from FireManager.
    package_name: Name of the extension package.

  Raises:
    PackageRegistrationError: The provided FireManager mixin is invalid.
  """
  if manager_cli_mixin is None:
    return

  if inspect.isclass(manager_cli_mixin):  # Non-classes don't have __mro__.
    # fire_manager is not imported for a subclass check to prevent a circular
    # import.
    mro_class_names = [
        a_class.__name__ for a_class in manager_cli_mixin.__mro__]
    is_invalid = (inspect.isabstract(manager_cli_mixin)
                  or "FireManager" not in mro_class_names)
  else:
    is_invalid = True

  if is_invalid:
    raise errors.PackageRegistrationError(
        "Provided FireManager mixin class is invalid. It must be None or a "
        "concrete class object which inherits from gazoo_device.FireManager.",
        package_name=package_name)


def _assert_subclasses(classes: Collection[Type[Any]],
                       parent: Type[Any],
                       package_name: str,
                       class_description: str,
                       allow_abstract: bool = False) -> None:
  """Raises an error if classes are not (concrete) subclasses of parent.

  Args:
    classes: Class objects to check.
    parent: The class object which should be the parent for all of classes.
    package_name: Name of the package being registered.
    class_description: Description of the classes being registered.
      For example, "primary device".
    allow_abstract: If False, all of classes must be concrete. Otherwise,
      abstract classes are allowed.

  Raises:
    PackageRegistrationError: One or more classes are invalid (not a class
      object, not a subclass of parent, or abstract when allow_abstract=False).
  """
  error_messages = []
  not_classes = [a_class for a_class in classes if not inspect.isclass(a_class)]
  if not_classes:
    error_messages.append(f"{not_classes} must be class objects.")

  not_subclasses = [  # issubclass() only works on class objects.
      a_class for a_class in classes
      if inspect.isclass(a_class) and not issubclass(a_class, parent)]
  if not_subclasses:
    error_messages.append(
        f"{not_subclasses} must be subclasses of {parent.__name__}.")

  if not allow_abstract:
    abstract_classes = [a_class for a_class in classes
                        if inspect.isabstract(a_class)]
    if abstract_classes:
      error_messages.append(f"{abstract_classes} must not be abstract.")

  if error_messages:
    raise errors.PackageRegistrationError(
        "Provided {} classes are invalid. {}".format(class_description,
                                                     " ".join(error_messages)),
        package_name=package_name)


def _assert_unique(names: Collection[str],
                   names_description: str,
                   classes_description: str,
                   package_name: str) -> None:
  """Raises an error if the names are not unique.

  Args:
    names: Class names to validate.
    names_description: Description of the class names.
    classes_description: Desciption of the classes.
    package_name: Name of the package being registered.

  Raises:
    PackageRegistrationError: Class names are not unique.
  """
  name_counts = collections.Counter(names)
  duplicate_names = [name for name, count in name_counts.items() if count > 1]
  if duplicate_names:
    raise errors.PackageRegistrationError(
        f"{names_description} {duplicate_names} are used by multiple "
        f"{classes_description}. {names_description} must be unique.",
        package_name=package_name)


def _raise_if_redefined(new_classes: Mapping[str, Type[Any]],
                        old_classes: Mapping[str, Type[Any]],
                        classes_description: str,
                        package_name: str) -> None:
  """Raises if the new class dictionary redefines items in the old one.

  Args:
    new_classes: New class name -> new class object mapping.
    old_classes: Known class name -> known class object mapping.
    classes_description: Human-readable description of the classes.
    package_name: Name of the package being registered.

  Raises:
    PackageRegistrationError: Some of the new classes share names with the
      already known classes.
  """
  redefined_names = list(set(old_classes.keys()) & set(new_classes.keys()))
  if redefined_names:
    offending = [new_classes[if_name] for if_name in redefined_names]
    redefined = [old_classes[if_name] for if_name in redefined_names]
    raise errors.PackageRegistrationError(
        f"New {classes_description} {offending} have same names "
        f"({redefined_names}) as existing {classes_description} {redefined}.",
        package_name=package_name)


def _copy_extensions():
  """Returns a backup copy of the current state of extensions."""
  return (extensions.capability_interfaces.copy(),
          extensions.capabilities.copy(),
          extensions.capability_flavors.copy(),
          extensions.auxiliary_devices.copy(),
          extensions.primary_devices.copy(),
          extensions.virtual_devices.copy(),
          extensions.communication_types.copy(),
          copy.deepcopy(extensions.detect_criteria),
          copy.deepcopy(extensions.package_info),
          extensions.keys.copy())


def _restore_extensions(backup_copy):
  """Restores the state of extensions from the backup copy."""
  (extensions.capability_interfaces,
   extensions.capabilities,
   extensions.capability_flavors,
   extensions.auxiliary_devices,
   extensions.primary_devices,
   extensions.virtual_devices,
   extensions.communication_types,
   extensions.detect_criteria,
   extensions.package_info,
   extensions.keys) = backup_copy



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Built-in GDM communication types.

A communication type identifies all possible connections and maps them to
transports, buttons, data framers, and line identifiers.
"""
import abc
import os
import types
from typing import Any, Collection, Dict, List, Optional, Union

from gazoo_device import custom_types
from gazoo_device import data_types
from gazoo_device import errors
from gazoo_device import extensions
from gazoo_device import gdm_logger
from gazoo_device.switchboard import data_framer
from gazoo_device.switchboard import ftdi_buttons
from gazoo_device.switchboard import line_identifier
from gazoo_device.switchboard.transports import adb_transport
from gazoo_device.switchboard.transports import jlink_transport
from gazoo_device.switchboard.transports import pigweed_rpc_transport
from gazoo_device.switchboard.transports import pty_transport
from gazoo_device.switchboard.transports import serial_transport
from gazoo_device.switchboard.transports import ssh_transport
from gazoo_device.switchboard.transports import transport_base
from gazoo_device.utility import adb_utils
from gazoo_device.utility import host_utils
from gazoo_device.utility import usb_config
from gazoo_device.utility import usb_utils

logger = gdm_logger.get_logger()

JLINK_COMMS_PRODUCT_NAME = "J-Link"
CP2104_COMMS_PRODUCT_NAME = "CP2104"
USB_COMMS_PRODUCT_NAME = "USB Single Serial"
NRF_DK_COMMS_ADDRESS_LINUX = "SEGGER_J-Link"
NRF_DK_EFR32_COMMS_ADDRESS_MAC = "tty.usbmodem"
EFR32_COMMS_ADDRESS_LINUX = "Silicon_Labs_J-Link"
ESP32_M5STACK_CP2104_COMMS_ADDRESS_LINUX = "Silicon_Labs_CP2104"
ESP32_M5STACK_CP2104_COMMS_ADDRESS_MAC = "tty.usbserial"
ESP32_M5STACK_CH9102F_COMMS_ADDRESS_LINUX = "1a86_USB_Single_Serial"


def _validate_comm_types(
    comm_types: Optional[Collection[str]] = None) -> Optional[Collection[str]]:
  """Validate that the communication types specified for detection are valid.

  A warning will be logged if a non-supported communication is included in
  the provided comm types.

  Args:
   comm_types: Specific communication types to use for detection.

  Returns:
    Lowercase supported communication types to use for detection.
  """
  if comm_types is None:
    return None

  lowercase_supported_types = [comm_type.lower() for comm_type in
                               extensions.communication_types.keys()]
  validated_comm_types = []
  invalid_comm_types = []
  for comm_type in comm_types:
    if comm_type.lower() in lowercase_supported_types:
      validated_comm_types.append(comm_type.lower())
    else:
      invalid_comm_types.append(comm_type)

  if invalid_comm_types:
    logger.warning(
        "Unknown communication types specified "
        f"{', '.join(invalid_comm_types)}. Known communication types "
        f"{', '.join(extensions.communication_types.keys())}.")

  return validated_comm_types


def detect_connections(
    static_ips: Optional[List[str]] = None,
    comm_types: Optional[Collection[str]] = None) -> Dict[str, List[str]]:
  """Detects all the communication addresses for the different devices.

  Args:
    static_ips: Static ip addresses.
    comm_types: Limit detection to specific communication types.

  Returns:
    Connections by connection class name from classes in this module
    and other registered classes in extensions.communication_types.
  """
  lowercase_comm_types = _validate_comm_types(comm_types)

  connections_dict = {}
  for comms_name, comms_class in extensions.communication_types.items():
    if lowercase_comm_types is not None:
      if comms_name.lower() not in lowercase_comm_types:
        logger.debug(
            f"Skipping detection for {comms_name} communication addresses.")
        continue

    logger.info(
        "\tdetecting potential {} communication addresses".format(comms_name))
    detection_method = comms_class.get_comms_addresses
    try:
      try:
        comms_addresses = detection_method(static_ips=static_ips)
      except TypeError:  # method does not accept static_ips
        comms_addresses = detection_method()
      connections_dict[comms_name] = comms_addresses
      logger.debug("Found {} potential communication addresses".format(
          len(comms_addresses)))

      # Verify ssh keys exist if ssh connections are detected
      if issubclass(comms_class, SshComms) and comms_addresses:
        missing_keys = []
        ssh_keys = [
            key_info for key_info in extensions.keys
            if key_info.type == data_types.KeyType.SSH
        ]
        for ssh_key in ssh_keys:
          try:
            host_utils.verify_key(ssh_key)
          except ValueError as err:  # Failed to set permissions on the key
            logger.warning(repr(err))
          except (errors.DownloadKeyError, FileNotFoundError, RuntimeError):
            missing_keys.append(ssh_key)
        if missing_keys:
          logger.warning("Found {} missing SSH keys:\n{}\n"
                         "Detection of SSH devices may not work correctly. "
                         "Run 'gdm download-keys'.".format(
                             len(missing_keys),
                             "\n".join(str(key) for key in missing_keys)))
    except Exception as err:  # pylint: disable=broad-except
      logger.warning(
          "Unable to detect {} communication addresses. Err: {!r}".format(
              comms_name, err))
  return connections_dict


def get_specific_serial_addresses(
    match_criteria: custom_types.MatchCriteria) -> List[str]:
  """Returns addresses for connections that match criteria.

  Excludes Android devices.
  Connections first excluded and then included.
  If there is no include regex, all instances that don't match exclude_regex
  are included.

  Dictionary entry: {'<key>':
  {'include_regex': <regex>, 'exclude_regex': '<regex>'}}

  Args:
    match_criteria: Criteria for finding a match.

  Returns:
    Addresses from UsbInfo instances that match criteria.
  """
  instances = usb_utils.find_matching_connections(match_criteria)
  accessible_instances = [
      instance for instance in instances
      if os.access(instance.address, os.R_OK | os.W_OK)
  ]
  inaccessible_addresses = [
      instance.address
      for instance in instances
      if instance not in accessible_instances
  ]
  if inaccessible_addresses:
    logger.warning(
        "No read/write permission for these serial address(es): {}".format(
            inaccessible_addresses))
  return [
      instance.address
      for instance in accessible_instances
      if instance.product_name not in usb_config.ANDROID_NAMES
  ]


class CommunicationType(abc.ABC):
  """Default settings for a communication type.

  Child classes should at minimum overwrite get_transport_list and
  get_comms_addresses.
  Child classes will likely overwrite init with their required args.
  """

  def __init__(self, comms_address: str) -> None:
    """Initiates args."""
    self.comms_address = comms_address

  def get_button_list(self) -> List[ftdi_buttons.FtdiButtons]:
    """Set ups the list of FTDI button instances as needed.

    For non-ftdi devices it just returns an empty list.

    Returns:
      FTDI button instances.
    """
    return []

  @classmethod
  @abc.abstractmethod
  def get_comms_addresses(cls) -> List[str]:
    """Detects the comms addresses that match this communication type.

    List can overlap those from other communication types.

    Returns:
      Potential communication addresses.
    """

  def get_data_framers(self,
                       num_transports: int) -> List[data_framer.DataFramer]:
    """Set up framers used to atomicize the raw output of the device.

    Deals with interwoven lines as well as tokenized output. Default
    is to use new line character to separate the lines.

    Args:
      num_transports: Number of declared transports.

    Returns:
      Data framers mapped to each transport.
    """
    return [data_framer.NewlineFramer()] * num_transports

  def get_identifier(self) -> line_identifier.AllUnknownIdentifier:
    """Setup identifiers used to distinguish loglines from responses.

    Defaults to labelling all as unknown.

    Returns:
      LineIdentifier: type of line identifier.
    """
    return line_identifier.AllUnknownIdentifier()

  def get_partial_line_timeout_list(self, num_transports: int) -> List[float]:
    """Time to wait before returning partial lines for a device.

    Args:
      num_transports: Number of declared transports.

    Returns:
      Timeouts for each transport.
    """
    return [.1] * num_transports

  def get_switchboard_kwargs(self) -> Dict[str, Any]:
    """Returns args needed to set up switchboard for this communication type.

    Returns:
      Dictionary of transport_list, identifier, framer_list, button_list.
    """
    transport_list = self.get_transport_list()
    num_transports = len(transport_list)
    return {
        "transport_list":
            transport_list,
        "identifier":
            self.get_identifier(),
        "framer_list":
            self.get_data_framers(num_transports),
        "button_list":
            self.get_button_list(),
        "partial_line_timeout_list":
            self.get_partial_line_timeout_list(num_transports),
    }

  @abc.abstractmethod
  def get_transport_list(self) -> List[transport_base.TransportBase]:
    """Setups up the transports which will communicate raw device io.

    Returns:
      list: Initiated transports.
    """


class AdbComms(CommunicationType):
  """Communication type for devices communicating over ADB.

  Commands/responses go over port 0 while logs are streamed over port 1.
  """

  def __init__(self,
               comms_address: str,
               log_cmd: str = "logcat -v threadtime",
               shell_cmd: str = "shell",
               event_log_cmd: Optional[str] = None) -> None:
    super().__init__(comms_address)
    self.log_cmd = log_cmd
    self.shell_cmd = shell_cmd
    self.event_log_cmd = event_log_cmd

  @classmethod
  def get_comms_addresses(cls) -> List[str]:
    """Returns ADB identifiers of available ADB devices."""
    return adb_utils.get_adb_devices()

  def get_transport_list(self) -> List[transport_base.TransportBase]:
    transports = [
        adb_transport.AdbTransport(
            comms_address=self.comms_address, command=self.shell_cmd),
        adb_transport.AdbTransport(
            comms_address=self.comms_address, command=self.log_cmd),
    ]
    if self.event_log_cmd:
      transports.append(
          adb_transport.AdbTransport(
              comms_address=self.comms_address, command=self.event_log_cmd))
    return transports

  def get_identifier(self) -> line_identifier.LineIdentifier:
    return line_identifier.PortLogIdentifier(log_ports=[1, 2])


class DockerComms(CommunicationType):
  """Communication type for VDL Docker devices."""

  def __init__(self,
               comms_address: str,
               secondary_comms_address: Optional[str] = None,
               secondary_log_pattern: Optional[str] = None) -> None:
    super().__init__(comms_address)
    self.secondary_comms_address = secondary_comms_address
    self.secondary_log_pattern = secondary_log_pattern

  @classmethod
  def get_comms_addresses(cls) -> List[str]:
    return host_utils.get_all_vdl_docker_connections()

  def get_transport_list(self) -> List[transport_base.TransportBase]:
    docker_exec_args = "exec -i {} /bin/bash".format(self.comms_address)
    docker_logs_args = "logs -f {}".format(self.comms_address)
    return [
        pty_transport.PtyTransport("docker", docker_exec_args),
        pty_transport.PtyTransport("docker", docker_logs_args)
    ]


class JlinkSerialComms(CommunicationType):
  """Communication type for JLINK and UART communication simultaneously.

  Serial is used for primary communication while jlink is used for log
  streaming.
  """

  @classmethod
  def get_comms_addresses(cls):
    return []

  def __init__(self, comms_address: str, baudrate: int = 115200) -> None:
    super().__init__(comms_address)
    self.secondary_address = usb_utils.get_serial_number_from_path(
        comms_address)
    self.baudrate = baudrate

  def get_transport_list(self) -> List[transport_base.TransportBase]:
    return [
        serial_transport.SerialTransport(
            comms_address=self.comms_address, baudrate=self.baudrate),
        jlink_transport.JLinkTransport(comms_address=self.secondary_address)
    ]

  def get_identifier(self) -> line_identifier.PortLogIdentifier:
    return line_identifier.PortLogIdentifier()


class PigweedSerialComms(CommunicationType):
  """Communication type for device communication over Pigweed RPC serial."""

  @classmethod
  def get_comms_addresses(cls) -> List[str]:
    include_product = [
        JLINK_COMMS_PRODUCT_NAME, CP2104_COMMS_PRODUCT_NAME,
        USB_COMMS_PRODUCT_NAME]
    include_address = [
        NRF_DK_COMMS_ADDRESS_LINUX, NRF_DK_EFR32_COMMS_ADDRESS_MAC,
        EFR32_COMMS_ADDRESS_LINUX, ESP32_M5STACK_CP2104_COMMS_ADDRESS_LINUX,
        ESP32_M5STACK_CP2104_COMMS_ADDRESS_MAC,
        ESP32_M5STACK_CH9102F_COMMS_ADDRESS_LINUX
    ]
    match_criteria = {
        "product_name": {
            "include_regex": "|".join(include_product)
        },
        "address": {
            "include_regex": "|".join(include_address)
        }
    }
    return get_specific_serial_addresses(match_criteria)

  def __init__(self,
               comms_address: str,
               protobufs: Collection[types.ModuleType],
               baudrate: int = serial_transport.DEFAULT_BAUDRATE):
    super().__init__(comms_address)
    self.protobufs = protobufs
    self.baudrate = baudrate

  def get_transport_list(self) -> List[transport_base.TransportBase]:
    protobuf_import_paths = [module.__name__ for module in self.protobufs]
    return [
        pigweed_rpc_transport.PigweedRPCTransport(
            comms_address=self.comms_address,
            protobuf_import_paths=protobuf_import_paths,
            baudrate=self.baudrate)
    ]


class PtyProcessComms(CommunicationType):
  """Communication type for communication with a process on the host machine.

  No detection provided.
  """

  @classmethod
  def get_comms_addresses(cls) -> List[str]:
    return host_utils.get_all_pty_process_directories()

  def __init__(self, comms_address: str) -> None:
    full_command_list = comms_address.split()
    comms_address = full_command_list[0]
    super().__init__(comms_address)
    self.args = " ".join(full_command_list[1:])

  def get_transport_list(self) -> List[pty_transport.PtyTransport]:
    return [pty_transport.PtyTransport(self.comms_address, args=self.args)]

  def get_identifier(self) -> line_identifier.AllUnknownIdentifier:
    return line_identifier.AllUnknownIdentifier()


class SshComms(CommunicationType):
  """Communication type for device communication over ssh.

  Login is handled by ssh-keys on machine (or device does not require
  password).

  Commands/responses are on port 0 while log streaming happens on port 1.
  """

  @classmethod
  def get_comms_addresses(cls, static_ips: List[str]) -> List[str]:
    return host_utils.get_all_ssh_ips(static_ips)

  def __init__(self,
               comms_address: str,
               log_cmd: str = "tail -F -n /var/log/messages",
               args: str = host_utils.DEFAULT_SSH_OPTIONS,
               key_info: Optional[data_types.KeyInfo] = None,
               username: str = "root") -> None:
    super().__init__(comms_address)
    self.log_cmd = log_cmd
    self.args = args
    self.username = username
    self.key_info = key_info

  def get_transport_list(self) -> List[ssh_transport.SSHTransport]:
    return [
        ssh_transport.SSHTransport(
            comms_address=self.comms_address,
            args=self.args,
            key_info=self.key_info,
            username=self.username),
        ssh_transport.SSHTransport(
            comms_address=self.comms_address,
            args=self.args,
            username=self.username,
            key_info=self.key_info,
            log_cmd=self.log_cmd)
    ]

  def get_identifier(self) -> line_identifier.PortLogIdentifier:
    return line_identifier.PortLogIdentifier(log_ports=[1])


class SnmpComms(CommunicationType):
  """Communication type for device communication over snmp protocol."""

  @classmethod
  def get_comms_addresses(cls, static_ips: List[str]) -> List[str]:
    return host_utils.get_all_snmp_ips(static_ips)

  def get_transport_list(self) -> List[transport_base.TransportBase]:
    # This comms type does not use Swtichboard so no transports exist.
    return []


class SerialComms(CommunicationType):
  """Communication type for device communication over basic serial."""

  def __init__(self,
               comms_address: str,
               log_line_regex: Optional[str] = None,
               baudrate: int = serial_transport.DEFAULT_BAUDRATE):

    super().__init__(comms_address)
    self.log_line_regex = log_line_regex
    self.baudrate = baudrate

  @classmethod
  def get_comms_addresses(cls) -> List[str]:
    exclude_products = [
        JLINK_COMMS_PRODUCT_NAME,
    ]
    exclude_product_ids = [
    ]

    match_criteria = {
        "product_name": {
            "exclude_regex": "|".join(exclude_products)
        },
        "product_id": {
            "exclude_regex": "|".join(exclude_product_ids)
        },
        "ftdi_interface": {
            "include_regex": "0|1"
        }
    }
    return get_specific_serial_addresses(match_criteria)

  def get_transport_list(self) -> List[serial_transport.SerialTransport]:
    return [
        serial_transport.SerialTransport(
            comms_address=self.comms_address, baudrate=self.baudrate)
    ]

  def get_identifier(
      self
  ) -> Union[line_identifier.AllUnknownIdentifier,
             line_identifier.RegexLogIdentifier]:
    if self.log_line_regex:
      return line_identifier.RegexLogIdentifier(self.log_line_regex)
    return super().get_identifier()


class UsbComms(CommunicationType):
  """Communication type for USB devices which don't leverage switchboard."""

  @classmethod
  def get_comms_addresses(cls) -> List[str]:
    devices = usb_utils.get_usb_devices_having_a_serial_number()
    return [d.serial_number for d in devices]

  def get_transport_list(self) -> List[transport_base.TransportBase]:
    return []


class YepkitComms(CommunicationType):
  """Communication type for yepkit devices.

  Currently supports just detection as communication does not yet work as
  expected.
  """

  @classmethod
  def get_comms_addresses(cls) -> List[str]:
    return host_utils.get_all_yepkit_serials()

  def get_transport_list(self) -> List[transport_base.TransportBase]:
    return []



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for patching Python fire library."""
import sys

import fire


def display(lines, out):
  """Display text to user.

  Args:
      lines (str): lines of text to display
      out (object): file object used by interpreter for output
  Note:
      original behavior: lines displayed in scrollable window, similar to
        "less" shell command.
      behavior with patch: lines printed to the console in standard format.
  """
  if not (out is sys.stdout or out is sys.stderr):
    out = sys.stdout
  text = '\n'.join(lines) + '\n'
  out.write(text)


def visible_members(component, class_attrs=None, verbose=False):
  """Returns a list of the members of the given component.

  Args:
      component (object): the component whose members to list.
      class_attrs (dict): determines how class members will be treated for
        visibility, defaulting to GetClassAttrsDict.
      verbose (bool): whether to include private members.

  Returns:
    list: of tuples (member_name, member) of all members of the component.

  Note:
      original behavior: component members are collected using
      inspect.getmembers.
      behavior with patch: component members are collected using
      _safely_get_members.
  """
  if isinstance(component, dict):
    members = component.items()
  else:
    members = _safely_get_members(component)

  # if class_attrs has not been provided, compute it.
  if class_attrs is None:
    # pytype: disable=module-attr
    try:
      class_attrs = fire.inspectutils.GetClassAttrsDict(component)
    except AttributeError:
      class_attrs = fire.completion.GetClassAttrsDict(component)
    # pytype: enable=module-attr

  return [(
      member_name, member
  ) for member_name, member in members if fire.completion.MemberVisible(
      component, member_name, member, class_attrs=class_attrs, verbose=verbose)]


def apply_patch():
  """Applies monkey patch by attaching custom functions to the Python Fire module."""
  fire.core.Display = display
  fire.completion.VisibleMembers = visible_members


def _safely_get_members(component):
  """Gather stable members of a component.

  Args:
      component (object): The component from which to get the members from.

  Returns:
      list: of stable component members as tuples (member name, member
      instance).

  Note:
      This function was designed to silence exceptions thrown by unstable
      devices when all
      device properties were previously accessed by inspect.getmembers.
  """
  members = []
  for attr_name in dir(component):
    try:
      attribute = getattr(component, attr_name)
      members.append((attr_name, attribute))
    except Exception:
      pass  # ignore members that raise exceptions

  return members



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gets information on all usb connected devices.


Returns a dictionary mapping a persistent address (adb or serial port) to usb
info for that address.

Note: usb_info_mac primarily relies on system_profile for full information.
"""
import copy
import json
import os
import subprocess
from typing import Dict, List, Any, Optional
from gazoo_device import gdm_logger
from gazoo_device.utility import usb_config
from serial.tools import list_ports

_DeviceDictType = Dict[str, Any]
_UsbInfoDict = Dict[str, usb_config.UsbInfo]
logger = gdm_logger.get_logger()

PROFILER_USB_DATA_TYPE = 'SPUSBDataType'
MAC_COMMAND = ['system_profiler', PROFILER_USB_DATA_TYPE, '-json']


def get_address_to_usb_info_dict():
  """Returns a standard address to usb info dict for mac.

  Returns:
      dict: of address to UsbInfo instance.

  Raises:
      RuntimeError: if system_profiler does not produce the correct output.
  """
  with open(os.devnull, 'w') as devnull:
    output = subprocess.check_output(MAC_COMMAND, stderr=devnull)
  output_dict = json.loads(output)
  usb_connections = output_dict[PROFILER_USB_DATA_TYPE]
  usb_info_dict = {}
  _convert_system_profiler_dict(usb_connections, usb_info_dict)
  return usb_info_dict


def _convert_system_profiler_dict(
    device_dicts: List[_DeviceDictType],
    converted_dicts: _UsbInfoDict,
    port_mapping: Optional[Dict[str, int]] = None,
    parent_device_dict: Optional[Dict[str, str]] = None) -> None:
  """Convert system profiler dict to the system agnostic form.

  Recursively walks the tree of connected USB devices.

  Args:
      device_dicts: dictionaries from system_profiler to convert to a
        generic form.
      converted_dicts: converted dictionaries will be stored here.
      port_mapping: mapping from USB device number (str) to Cambrionix
        port number (int).
      parent_device_dict: dictionary of parent hub (Cambrionix).
  """
  for device_dict in device_dicts:
    if (device_dict.get('manufacturer') == 'cambrionix' and
        device_dict.get('_name') in usb_config.CAMBRIONIX_PORT_MAP and
        _get_vendor_product_id(device_dict) not in
        usb_config.CAMBRIONIX_VENDOR_PRODUCT_ID):
      children = _create_device_entries_for_cambrionix(device_dict,
                                                       device_dict['_name'])
      converted_dicts.update(children)
    else:
      if 'serial_num' in device_dict:
        entry = _create_device_entry(
            device_dict, port_mapping, parent_device_dict)
        converted_dicts.update(entry)
      _convert_system_profiler_dict(
          device_dict.get('_items', []),
          converted_dicts,
          port_mapping,
          parent_device_dict)


def _add_child_addresses(children_dicts):
  """Add child addresses to cambrionix entry and cambrionix to children."""
  child_addresses = list(children_dicts.keys())
  cambrionix_address = [
      key for key, value in children_dicts.items()
      if value.product_name in usb_config.CAMBRIONIX_NAMES
  ][0]
  children_dicts[cambrionix_address].child_addresses = child_addresses
  for address in child_addresses:
    if address != cambrionix_address:
      children_dicts[address].usb_hub_address = cambrionix_address


def _create_device_entries_for_cambrionix(device_dict, cambrionix_model):
  """Converts device dict for cambrionix into individual entries for children."""
  port_mapping = usb_config.CAMBRIONIX_PORT_MAP[cambrionix_model]
  children_dicts = {}
  # Typically the entries are in a _items list, however at least for the
  # PS15-USB3 hub the cambrionix device is directly in the 'entry'
  for entry in device_dict.get('_items', []):
    if '_items' in entry:
      entry_device_dict = entry.get('_items')
    else:
      entry_device_dict = [entry]
    _convert_system_profiler_dict(
        entry_device_dict,
        children_dicts,
        port_mapping,
        device_dict)
  if children_dicts:
    _add_child_addresses(children_dicts)
  return children_dicts


def _create_device_entry(
    device_dict: _DeviceDictType,
    port_mapping: Optional[Dict[str, int]] = None,
    parent_device_dict: Optional[_DeviceDictType] = None) -> _UsbInfoDict:
  """Creates a standard usb_info entry for a device.

  Args:
   device_dict: device info from system profiler.
   port_mapping: map of port identifiers to port numbers.
   parent_device_dict: dictionary of parent hub (Cambrionix).

  Returns:
    dict: UsbInfo dictionary entries stored by address as the key.
  """
  return_dict = {}
  usb_entry = usb_config.UsbInfo()
  usb_entry.serial_number = device_dict['serial_num']
  usb_entry.manufacturer = device_dict.get('manufacturer', '')  # can be empty
  usb_entry.product_id = _get_product_id(device_dict)
  usb_entry.vendor_id = _get_vendor_id(device_dict)
  # Product names are reported inconsistently between gLinuxes and Macs; either
  # undescores or spaces are used. Convert all underscores to spaces as a
  # workaround.
  usb_entry.product_name = device_dict['_name'].replace('_', ' ')
  usb_entry.usb_hub_port = _get_port_number(
      device_dict, port_mapping, parent_device_dict)
  address = _get_entry_address(usb_entry.serial_number)
  if 'Media' in device_dict:  # mounted device

    if 'volumes' in device_dict['Media'][0]:
      volume = device_dict['Media'][0]['volumes'][0]
      address = '/dev/' + volume['bsd_name']
      usb_entry.disk = address
  if usb_entry.product_name in usb_config.ANDROID_NAMES:
    usb_entry.address = usb_entry.serial_number  # android serial
    return_dict[usb_entry.address] = usb_entry
  else:
    usb_entry.address = address
    if os.path.exists(usb_entry.address):
      return_dict[usb_entry.address] = usb_entry
  return return_dict


def _get_entry_address(device_serial: str) -> str:
  """Returns serial port path given device serial number."""
  comports = list_ports.comports()
  for com_port in comports:
    if com_port.serial_number == device_serial:
      return com_port.device.replace('/cu.', '/tty.')
  return ''


def _get_product_id(device_dict):
  """Converts 0x1234 to 1234."""
  return device_dict['product_id'].split('x')[-1]


def _get_vendor_id(device_dict):
  """Converts '0x1234 Serial Interational' to 1234."""
  return device_dict['vendor_id'].split()[0].split('x')[-1]


def _get_vendor_product_id(device_dict):
  """Returns a string with '<vendor_id>/<product_id>'."""
  return f'{_get_vendor_id(device_dict)}/{_get_product_id(device_dict)}'


def _get_port_number(
    device_dict: _DeviceDictType,
    port_mapping: Optional[Dict[str, int]] = None,
    parent_device_dict: Optional[_DeviceDictType] = None) -> Optional[int]:
  """Last two digits correspond to the port number.

  Args:
    device_dict: device info from system profiler.
    port_mapping: map of port identifiers to port numbers.
    parent_device_dict: dictionary of parent hub (Cambrionix).

  Returns:
     int: port number or None if no cambrionix port applicable..

  Examples:
      "0x14443000 / 10" -> port identifier is "43" -> port 5 on PP15S.
      "0x14313400" -> port identifier is "34" -> port 8 on PP15S.
  """
  if not port_mapping:  # not connected to a usb hub
    return None
  if is_cambrionix(device_dict):
    return None
  if parent_device_dict:
    index = _get_cambrionix_port_using_parent_hub(
        device_dict, parent_device_dict)
  else:
    # get the appropriate digits
    stripped_location_id = _rstrip_location_id(device_dict)
    index = '{}.{}'.format(stripped_location_id[-2], stripped_location_id[-1])
  port = port_mapping[index]
  return port


def _get_cambrionix_port_using_parent_hub(
    device_dict: _DeviceDictType,
    parent_device_dict: _DeviceDictType) -> str:
  """Returns device Cambrionix port using the parent (Cambrionix) dict.

  Args:
    device_dict: device info from system profiler.
    parent_device_dict: dictionary of parent hub (Cambrionix).

  Returns:
     str: Cambrionix port number in port_map dict format.

  Examples:
     Device:                       Location ID:
     Cambrionix                  = 0x146  <--- length = 5
     Child Cambrionix Hub        = 0x1464  <--- length = 6
                                        ^--- location_id[5]
     Child Device                = 0x14642
                                        ^--- (4.2 = port 1)
     Generic Hub                 = 0x14643
                                        ^--- (4.3 = port 2)
     Child Device of Generic Hub = 0x146431
                                        ^--- (4.3 = port 2)
  """
  first_loc_id_idx = len(_rstrip_location_id(parent_device_dict))
  second_loc_id_idx = first_loc_id_idx + 1
  device_loc_id = _rstrip_location_id(device_dict)
  return '{}.{}'.format(device_loc_id[first_loc_id_idx],
                        device_loc_id[second_loc_id_idx])


def is_cambrionix(device_dict):
  """Device dict is a cambrionix entry."""
  return device_dict.get('_name') in usb_config.CAMBRIONIX_NAMES


def _rstrip_location_id(device_dict):
  """Strips location_id down to base hex value with no trailing zeroes."""
  return device_dict['location_id'].split()[0].rstrip('0')



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Dimmable Light endpoint.

A Dimmable Light is a lighting device that is capable of being switched on or
off and the intensity of its light adjusted by means of a bound controller
device such as a Dimmer Switch or a Color Dimmer Switch.

This endpoint module corresponds to the
"Dimmable Light" Matter device type which defines required and optional clusters
in the Matter spec 1.0.

A Dimmable light endpoint houses an instance of Dimmable Light Matter device
type with the following defined clusters on this endpoint:

The required clusters for this endpoint: Scenes, OnOff, Level and Groups.
The optional clusters for this endpoint: Occupancy.
"""
from gazoo_device import decorators
from gazoo_device.capabilities.matter_clusters import level_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import occupancy_pw_rpc
from gazoo_device.capabilities.matter_clusters import on_off_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import dimmable_light_base


class DimmableLightEndpoint(dimmable_light_base.DimmableLightBase):
  """Matter Dimmable light endpoint."""

  @decorators.CapabilityDecorator(occupancy_pw_rpc.OccupancyClusterPwRpc)
  def occupancy(self) -> occupancy_pw_rpc.OccupancyClusterPwRpc:
    """Matter Occupancy cluster instance."""
    return self.cluster_lazy_init(occupancy_pw_rpc.OccupancyClusterPwRpc)

  @decorators.CapabilityDecorator(on_off_pw_rpc.OnOffClusterPwRpc)
  def on_off(self) -> on_off_pw_rpc.OnOffClusterPwRpc:
    """Matter OnOff cluster instance."""
    return self.cluster_lazy_init(on_off_pw_rpc.OnOffClusterPwRpc)

  @decorators.CapabilityDecorator(level_control_pw_rpc.LevelControlClusterPwRpc)
  def level(self) -> level_control_pw_rpc.LevelControlClusterPwRpc:
    """Matter Level Control cluster instance."""
    return self.cluster_lazy_init(level_control_pw_rpc.LevelControlClusterPwRpc)

  # TODO(b/209362086) Add the below clusters
  # def scenes(self):
  #   """Matter Scenes cluster instance."""
  #   # TODO(b/206741536): Implement Scenes cluster.
  #   raise NotImplementedError("Scenes cluster isn't implemented.")

  # def groups(self):
  #   """Matter Groups cluster instance."""
  #   # TODO(b/206741536): Implement Groups cluster.
  #   raise NotImplementedError("Groups cluster isn't implemented.")
